from polecat.db.schema.utils import Auto
from polecat.utils import to_tuple

from ..utils.stringcase import camelcase, snakecase
from .exceptions import InvalidModelDataError


class Field:
    query_resolvers = ()
    mutation_resolvers = ()
    hooks = []
    graphql_post_build_hooks = []

    def __init__(self, query_resolvers=None, mutation_resolvers=None,
                 omit=None, hooks=None):
        self.query_resolvers = to_tuple(query_resolvers or self.query_resolvers)
        self.mutation_resolvers = to_tuple(mutation_resolvers or self.mutation_resolvers)
        self.omit = omit or 0
        self.hooks = hooks or self.hooks

    def __repr__(self):
        name = getattr(self, 'name', '?')
        return f'<{self.__class__.__name__} name="{name}">'

    def prepare(self, model):
        return self

    def use_get_query(self):
        return False

    def from_incoming(self, instance, value):
        return value

    def to_outgoing(self, instance, value):
        return value


class MutableField(Field):
    def __init__(self, *args, null=False, unique=False, default=None,
                 primary_key=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.null = null
        self.unique = unique
        self.default = default
        self.primary_key = primary_key

    def use_get_query(self):
        return self.primary_key or self.unique


class TextField(MutableField):
    def __init__(self, *args, length=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.length = length


class EmailField(TextField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, length=255, **kwargs)


class PhoneField(TextField):
    pass


class PasswordField(TextField):
    pass


class BoolField(MutableField):
    pass


class IntField(MutableField):
    pass


class SerialField(MutableField):
    pass


class FloatField(MutableField):
    pass


class DateField(MutableField):
    pass


class DatetimeField(MutableField):
    pass


class UUIDField(MutableField):
    def __init__(self, *args, auto=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.auto = auto


class JSONField(MutableField):
    pass


class PointField(MutableField):
    pass


class GCSPointField(PointField):
    pass


class ReverseField(Field):
    def __init__(self, other, related_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.other = other
        self.related_name = related_name

    def from_incoming(self, instance, value):
        return [
            self.other(**{self.related_name: instance, **v})
            for v in value
        ]

    def to_outgoing(self, instance, value):
        return [
            self._model_to_dict(v)
            for v in (value or [])
        ]

    def _model_to_dict(self, model):
        # TODO: Ugh, no. Must fix.
        values = {}
        for field_name, field in model.Meta.fields.items():
            if not hasattr(model, field_name):
                continue
            if field_name == self.related_name:
                continue
            values[field_name] = field.to_outgoing(
                model,
                getattr(model, field_name)
            )
        return values


class RelatedField(MutableField):
    def __init__(self, other, related_name=Auto, on_delete=None, *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.other = other
        self.related_name = related_name
        self.on_delete = on_delete

    def prepare(self, model):
        if self.other == 'self':
            self.other = model
        self.add_related_field(model)
        return self

    def add_related_field(self, model):
        if self.related_name:
            related_name = self.make_related_name(
                self.related_name,
                model,
                self.other
            )
            if hasattr(self.other, related_name):
                # TODO: Better error.
                raise AttributeError('Related field name clash')
            # TODO: This whole process of adding an extra field to a
            # model is pretty awful. This should be handled somewhere
            # else.
            field = ReverseField(model, related_name=self.name)
            field.name = related_name  # TODO: Hmmm, not sure.
            field.cc_name = camelcase(related_name)
            self.other.Meta.fields[field.name] = field  # TODO: Ugh.
            self.other.Meta.cc_fields[field.cc_name] = field
            # TODO: Potentially information losing: now we don't know
            # that Auto was originally set.
            self.related_name = related_name

    def make_related_name(self, related_name, from_model, to_model):
        if related_name == Auto:
            # TODO: Depends on `name` being added after model
            # construction, not sure I like that.
            # TODO: Better pluralize? Could do, but it's typically
            # expensive.
            return f'{snakecase(from_model.Meta.plural)}_by_{self.name}'
        else:
            return related_name

    def from_incoming(self, instance, value):
        if isinstance(value, self.other):
            return value
        else:
            try:
                value = {'id': int(value)}
            except TypeError:
                pass
            try:
                return self.other(**value)
            except TypeError:
                raise InvalidModelDataError(self.other, value)

    def to_outgoing(self, instance, value):
        from .model import Model
        if isinstance(value, Model):
            if getattr(value, 'id', None) is not None:
                value = value.id
            else:
                value = self._model_to_dict(value)
        return value

    def _model_to_dict(self, model):
        # TODO: Ugh, no. Must fix.
        values = {}
        for field_name, field in model.Meta.fields.items():
            if not hasattr(model, field_name):
                continue
            if field_name == self.related_name:
                continue
            values[field_name] = field.to_outgoing(
                model,
                getattr(model, field_name)
            )
        return values


class ComputedField(Field):
    def __init__(self, function, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.function = function



import logging
from functools import partial

import ujson as json
from graphql.type import (GraphQLArgument, GraphQLBoolean, GraphQLEnumType,
                          GraphQLEnumValue, GraphQLField, GraphQLFloat,
                          GraphQLInputField, GraphQLInputObjectType,
                          GraphQLInt, GraphQLInterfaceType, GraphQLList,
                          GraphQLNonNull, GraphQLObjectType, GraphQLSchema,
                          GraphQLString)

from ..model import field, omit
from ..utils import add_attribute, capitalize
from .input import Input
from .registry import (FieldMetaclass, graphql_create_input_registry,
                       graphql_reverse_input_registry, graphql_type_registry,
                       graphql_update_input_registry)
from .type import GraphQLDatetime, GraphQLUUID, GraphQLDate

__all__ = ('Field', 'StringField', 'IntField', 'RelatedField')

logger = logging.getLogger(__name__)


class Field(metaclass=FieldMetaclass):
    graphql_type = None
    sources = ()

    def __init__(self, model, model_field, input=False, registry=None):
        self.model = model
        self.model_field = model_field
        self.input = input
        self.registry = registry or graphql_type_registry

    def make_graphql_field(self, name=None, optional=False):
        # TODO: Clean this up a bit.
        if self.input:
            # TODO: Use `optional` to remove required attributes. Used
            # for patches.
            self.graphql_field = add_attribute(
                GraphQLInputField(
                    self.get_graphql_type(),
                    description=self.get_description()
                ), '_field', self
            )
        else:
            self.graphql_field = add_attribute(
                GraphQLField(
                    self.get_graphql_type(),
                    resolve=self.get_resolver(),
                    description=self.get_description()
                ), '_field', self
            )
        return self.graphql_field

    def get_graphql_type(self):
        return self.graphql_type

    def get_resolver(self):
        return self.default_resolver

    def get_description(self):
        return ''

    def default_resolver(self, obj, path):
        return obj[self.model_field.name]

    def from_input(self, input, graphql_type):
        name = self.model_field.cc_name
        if name in input:
            change = {
                self.model_field.name: input[name]
            }
        else:
            change = None
        return change, None


class StringField(Field):
    graphql_type = GraphQLString
    sources = (field.TextField,)


class DateField(Field):
    graphql_type = GraphQLDate
    sources = (field.DateField,)


class DatetimeField(Field):
    graphql_type = GraphQLDatetime
    sources = (field.DatetimeField,)


class UUIDField(Field):
    graphql_type = GraphQLUUID
    sources = (field.UUIDField,)


class JSONField(Field):
    graphql_type = GraphQLString
    sources = (field.JSONField,)

    # TODO: Not sure this is the right way to do this.
    def default_resolver(self, obj, path):
        return json.dumps(super().default_resolver(obj, path))


class PointField(Field):
    graphql_type = GraphQLList(GraphQLFloat)
    sources = (field.PointField,)


class BoolField(Field):
    graphql_type = GraphQLBoolean
    sources = (field.BoolField,)


class IntField(Field):
    graphql_type = GraphQLInt
    sources = (field.IntField,)


class SerialField(Field):
    graphql_type = GraphQLInt
    sources = (field.SerialField,)


class FloatField(Field):
    graphql_type = GraphQLFloat
    sources = (field.FloatField,)


class RelatedField(Field):
    sources = (field.RelatedField,)

    # TODO: Should maybe have an extra function to avoid even getting
    # here?  Like `should_build`?
    def make_graphql_field(self, name=None, optional=False):
        if self.model_field.other.Meta.omit == omit.ALL:
            return None
        else:
            return super().make_graphql_field(name, optional)

    def get_graphql_type(self, registry=None):
        # TODO: I really need to be able to create and update with an
        # ID first and foremost, hence this change. This is so bad.
        registry = registry or self.registry
        if registry == graphql_update_input_registry or registry == graphql_reverse_input_registry or registry == graphql_create_input_registry:
            return GraphQLInt
        try:
            return registry[self.model_field.other]
        except KeyError:
            raise KeyError(f'unknown GraphQL type {self.model_field.other.Meta.name}')

    def from_input(self, input, graphql_type):
        # TODO: Nesting.
        return super().from_input(input, graphql_type)


class ReverseField(RelatedField):
    sources = (field.ReverseField,)

    def make_graphql_field(self, builder):
        if self.input:
            name = f'{self.model.Meta.name}{capitalize(self.model_field.cc_name)}Input'
            type = graphql_reverse_input_registry.get(name, None)
            if not type:
                logger.debug(f'Building reverse input for model {name} coming from {self.model.Meta.name}.{self.model_field.name}')
                # TODO: Noooooooo. Use the buidler getting passed in
                # to provide it.
                type = add_attribute(
                    GraphQLInputObjectType(
                        name=name,
                        fields=partial(self.build_reverse_fields, builder)
                    ), '_model', self.model
                )
                graphql_reverse_input_registry[name] = type
            return add_attribute(
                GraphQLInputField(
                    type,
                    description=self.get_description()
                ), '_field', self
            )
        else:
            return super().make_graphql_field()

    def build_reverse_fields(self, builder):
        from .schema import ReverseModelInputBuilder
        return {
            'create': GraphQLList(
                ReverseModelInputBuilder(builder).build(
                    self.model_field.other,
                    self.model_field.other.Meta.fields[self.model_field.related_name]
                )
            ),
            'delete': GraphQLList(builder.delete_type)
        }

    def get_graphql_type(self, registry=None):
        return GraphQLList(super().get_graphql_type(registry))

    def from_input(self, input, graphql_type):
        name = self.model_field.cc_name
        if name in input:
            # TODO: Nesting.
            # TODO: Is this a bit inefficient? All the conditionls?
            value = input[name] or {}
            delete = value.get('delete', set())
            change = value.get('create', None)
            if change:
                sub_type = graphql_type.fields[self.model_field.cc_name].type.fields['create'].type.of_type
                new_change = []
                for sub_args in change:
                    sub_input = Input(sub_type, sub_args)
                    sub_input.merge_delete(sub_input.delete, delete)
                    new_change.append(sub_input.change)
                change = {
                    self.model_field.name: new_change
                }
            if not len(delete):
                delete = None
        else:
            change, delete = None, None
        return change, delete



from . import Create


class Update(Create):
    def __init__(self, version, endpoint, id, data={}, replace_m2ms=True, **kwargs):
        super().__init__(version, endpoint, data, **kwargs)
        self._id = id
        self._replace_m2ms = replace_m2ms
        self._update = True

    def get_main_sql(self):
        table = self.endpoint['table']
        field_names, values = self.get_fields_values()
        fv_str = ', '.join(f'{f} = %s' for f in field_names)
        if not fv_str:
            return ''
        sql = f'UPDATE {table} SET {fv_str} WHERE id = {self._id} RETURNING id'
        return sql

    # TODO: Should this just be the create one?
    def get_m2m_sql(self):
        sql = super().get_m2m_sql()
        # for field_name, m2m_table, values in self._m2ms:
        #     if self._replace_m2ms:
        #         sql.append(f'DELETE FROM {m2m_table} WHERE left_id = {self._id}')
        #     if values:
        #         self._values.extend(values)
        #         v_str = ','.join(f'({self._id}, %s)' for v in values)
        #         sql.append(f'INSERT INTO {m2m_table}(left_id, right_id) VALUES {v_str}')
        sql.append(f'SELECT {self._id}')
        return sql



from .column import *  # noqa
from .role import *  # noqa
from .schema import *  # noqa
from .table import *  # noqa
from .utils import Auto  # noqa
from .variable import *  # noqa

CASCADE = 'CASCADE'



import click

from .feedback import HaloFeedback
from .main import main

__all__ = ('upload', 'upload_bundle')


@main.command()
@click.option('--source')
@click.option('--version')
@click.pass_context
def upload(ctx, source, version):
    from ..deploy.aws.upload import upload as aws_upload
    project = ctx.obj['project']
    bucket = ctx.obj['bucket']
    aws_upload(project, bucket, source, version, feedback=HaloFeedback())


@main.command()
@click.option('--source')
@click.option('--version')
@click.pass_context
def upload_bundle(ctx, source, version):
    from ..deploy.aws.upload import upload_bundle as aws_upload_bundle
    project = ctx.obj['project']
    bucket = ctx.obj['bucket']
    aws_upload_bundle(project, bucket, source, version, feedback=HaloFeedback())



from functools import partial
from uuid import uuid4

from factory import LazyFunction, declarations
from factory.fuzzy import FuzzyDateTime, FuzzyInteger, FuzzyText, FuzzyAttribute
from polecat.db.sql.postgres import Point
from polecat.utils import timezone

from ...core.registry import MappedRegistry, RegistryMetaclass
from ...model import field as mf

MappedRegistry('factory_field_registry')


class FactoryWrapper(declarations._FactoryWrapper):
    def __init__(self, factory_or_path):
        self.factory = None
        self.module = self.name = ''
        self.func = None
        if isinstance(factory_or_path, type):
            self.factory = factory_or_path
        elif callable(factory_or_path):
            self.func = factory_or_path
        else:
            if not (isinstance(factory_or_path, str) and '.' in factory_or_path):
                raise ValueError(
                    "A factory= argument must receive either a class "
                    "or the fully qualified path to a Factory subclass; got "
                    "%r instead." % factory_or_path)
            self.module, self.name = factory_or_path.rsplit('.', 1)

    def get(self):
        if self.factory is None:
            if self.func:
                self.factory = self.func()
        return super().get()


class SubFactory(declarations.SubFactory):
    def __init__(self, factory, **kwargs):
        super(declarations.SubFactory, self).__init__(**kwargs)
        self.factory_wrapper = FactoryWrapper(factory)


class Field(metaclass=RegistryMetaclass):
    _registry = 'factory_field_registry'
    _registry_base = 'Field'

    @classmethod
    def get_declaration(self, model, model_field, factory):
        raise NotImplemented


class TextField(Field):
    sources = (mf.TextField,)

    @classmethod
    def get_declaration(self, model, model_field, factory):
        kwargs = {}
        if model_field.length is not None:
            kwargs['length'] = model_field.length
        return FuzzyText(**kwargs)


class NumberField(Field):
    sources = (mf.IntField,)

    @classmethod
    def get_declaration(self, model, model_field, factory):
        # TODO: 0 isn't good.
        return FuzzyInteger(0)


class PointField(Field):
    sources = (mf.PointField, mf.GCSPointField)

    @classmethod
    def get_declaration(self, model, model_field, factory):
        # TODO: 0 isn't good.
        return LazyFunction(lambda: Point(0, 0))


class Date(Field):
    sources = (mf.DateField,)

    @classmethod
    def get_declaration(self, model, model_field, factory):
        return FuzzyDateTime(timezone.now().date())


class DatetimeField(Field):
    sources = (mf.DatetimeField,)

    @classmethod
    def get_declaration(self, model, model_field, factory):
        return FuzzyDateTime(timezone.now())


class BoolField(Field):
    sources = (mf.BoolField,)

    @classmethod
    def get_declaration(self, model, model_field, factory):
        # TODO
        return FuzzyAttribute(lambda: True)


class FloatField(Field):
    sources = (mf.FloatField,)

    @classmethod
    def get_declaration(self, model, model_field, factory):
        # TODO
        return FuzzyAttribute(lambda: 0.0)


class UUIDField(Field):
    sources = (mf.UUIDField,)

    @classmethod
    def get_declaration(self, model, model_field, factory):
        # TODO
        return FuzzyAttribute(lambda: uuid4())


class JSONField(Field):
    sources = (mf.JSONField,)

    @classmethod
    def get_declaration(self, model, model_field, factory):
        # TODO
        return FuzzyAttribute(lambda: '{}')


class RelatedField(Field):
    sources = (mf.RelatedField,)

    @classmethod
    def get_declaration(self, model, model_field, factory):
        if model_field.other != model:
            return SubFactory(partial(factory.get, model_field.other))
        else:
            return None



from .as_ import As
from .expression import Expression
from .select import Select
from .subquery import Subquery
from .union import Union


class SubrelationOverride(Expression):
    # TODO: Uh oh.
    override_counter = 0

    def __init__(self, expression, overrides, root_strategy):
        self.expression = expression
        self.overrides = overrides
        # TODO: This isn't nice.
        self.root_strategy = root_strategy

    def to_sql(self):
        return self.expression.to_sql()

    def get_subrelation(self, name):
        if name in self.overrides:
            override_queries = self.overrides[name]
            if len(override_queries) > 1:
                union = Union()
                for query in override_queries:
                    # TODO: What if it doesn't exist?
                    alias = self.root_strategy.get_query_alias(query)
                    union.add_expression(Select(alias))
                return As(Subquery(union), self.create_alias_name_for_override())
            else:
                return self.root_strategy.get_query_alias(override_queries[0])
        else:
            return self.expression.get_subrelation(name)

    def get_column(self, name):
        # TODO: Also bad design, but also what's the difference
        # between this and the above? Looks like the above just does
        # 'this.related_table'.
        return self.expression.get_column(name)

    def has_column(self, name):
        # TODO: Also bad design, but also what's the difference
        # between this and the above? Looks like the above just does
        # 'this.related_table'.
        return self.expression.has_column(name)

    def push_selection(self, selection=None):
        self.expression.push_selection(selection)

    def create_alias_name_for_override(self):
        alias_name = f'o{self.override_counter}'
        self.override_counter += 1
        return alias_name



import click  # noqa

from polecat.db.migration import migrate as db_migrate
from polecat.model.db.migrate import make_migrations as db_make_migrations
from polecat.project.project import load_project

from .main import main

__all__ = ('migrate', 'empty_migration', 'make_migrations')


@main.command()
def migrate():
    project = load_project()
    project.prepare()
    db_migrate(
        apps=project.apps,
        migration_paths=(project.path,)
    )


@main.command()
def empty_migration():
    pass


@main.command()
def make_migrations():
    project = load_project()
    project.prepare()
    db_make_migrations(migration_paths=(project.path,))



import click

from .feedback import HaloFeedback
from .main import main

__all__ = ('create',)


@main.group()
def certificate():
    pass


@certificate.command()
@click.argument('domain', nargs=-1)
@click.option('--name', '-n')
@click.pass_context
def create(ctx, domain, name):
    from polecat.deploy.aws.certificate import create_certificate
    create_certificate(domain, name, feedback=HaloFeedback())


@certificate.command()
@click.argument('domain')
@click.pass_context
def wait(ctx, domain):
    from polecat.deploy.aws.certificate import wait_certificate
    wait_certificate(domain, feedback=HaloFeedback())



import click

from polecat_feedback.summary import ListSummary
from termcolor import colored

from .feedback import HaloFeedback, cli_feedback
from .main import main

__all__ = ('create_instance', 'delete_instance', 'list_instances',
           'create', 'list_dbs', 'delete')


@main.group()
def db():
    pass


@db.command()
@click.option('--url')
@click.pass_context
def create_instance(ctx, url):
    from ..deploy.aws.db import create_instance as aws_create_instance
    aws_create_instance(url=url, feedback=HaloFeedback())


@db.command()
@click.pass_context
@cli_feedback('List database instances')
def list_instances(ctx, feedback=None):
    from ..deploy.aws.db import list_instances as aws_list_instances
    instances = aws_list_instances(feedback=feedback)
    # instances = ['blahdy-blah', 'hello-world']
    feedback.add_summary(
        ListSummary('Database instances', instances)
    )
    # if len(instances) == 0:
    #     print(colored('    No instances', 'grey'))
    # else:
    #     for inst in instances:
    #         print(f'    {inst}')


@db.command()
@click.argument('name')
@click.pass_context
def delete_instance(ctx, name):
    from ..deploy.aws.db import delete_instance as aws_delete_instance
    aws_delete_instance(name, feedback=HaloFeedback())


@db.command()
@click.argument('deployment')
@click.option('--instance')
@click.pass_context
def create(ctx, deployment, instance):
    from ..deploy.aws.db import create_db as aws_create_db
    project = ctx.obj['project']
    aws_create_db(project, deployment, instance_name=instance, feedback=HaloFeedback())


@db.command('list')
@click.argument('deployment')
@click.pass_context
def list_dbs(ctx, deployment):
    from ..deploy.aws.db import list_dbs as aws_list_dbs
    project = ctx.obj['project']
    dbs = aws_list_dbs(project, deployment, feedback=HaloFeedback())
    if len(dbs) == 0:
        print(colored('    No databases', 'grey'))
    else:
        for db in dbs:
            print(f'    {db}')


@db.command()
@click.argument('deployment')
@click.pass_context
def delete(ctx, deployment):
    from ..deploy.aws.db import delete_db as aws_delete_db
    project = ctx.obj['project']
    aws_delete_db(project, deployment, feedback=HaloFeedback())



import re

CAMELCASE_PROG = re.compile(r'_([a-z])')
SNAKECASE_PROG = re.compile(r'([A-Z]+)')


def camelcase(value):
    return (
        value[0].lower() +
        CAMELCASE_PROG.sub(lambda m: m.group(1).upper(), value[1:])
    )


def pascalcase(value):
    return (
        value[0].upper() +
        CAMELCASE_PROG.sub(lambda m: m.group(1).upper(), value[1:])
    )


def snakecase(value):
    # TODO: Needs testing.
    def do_sub(m):
        v = m.group(1)
        r = '_' if m.start() else ''
        if len(v) > 1:
            r += v[:-1].lower() + '_' + v[-1].lower()
        else:
            r += v.lower()
        return r
    return SNAKECASE_PROG.sub(do_sub, value)



from graphql_server import (default_format_error, encode_execution_results,
                            run_http_query)

from ..project.handler import Handler
from .schema import build_graphql_schema


class GraphqlAPI(Handler):
    def prepare(self):
        self.schema = build_graphql_schema()

    def match(self, event):
        # TODO: Use regex
        return event.is_http() and \
           event.request.method == 'POST' and \
           (event.request.path == '/graphql' or
            event.request.path == '/graphql/')

    async def handle_event(self, event):
        # TODO: There's no sync version of this for some reason...
        result = await run_http_query(
            self.schema,
            event.request.method.lower(),
            event.request.json,
            # query_data=None,
            # batch_enabled=False,
            # catch=False,
            context_value={
                'event': event,
                'session': event.session
            }
        )
        is_batch = False  # TODO: How to handle this?
        result = encode_execution_results(
            result[0],
            default_format_error,
            is_batch,
            lambda d: d  # pass-through
        )
        # TODO: This is a little ugly.
        if result[0]['errors']:
            result[0]['errors'] = tuple(map(format_error, result[0]['errors']))
        return result


def format_error(error):
    original_error = getattr(error, 'original_error', None)
    if original_error:
        return str(original_error)
    else:
        # TODO: If in debug mode, include path: str(error)
        try:
            return error.message
        except AttributeError:
            return str(error)



def unparse_url(info):
    scheme = info['scheme']
    auth = info.get('username', '')
    if auth:
        password = info.get('password', None)
        if password:
            auth += f':{password}'
        auth += '@'
    host = info['host']
    dbname = info['dbname']
    return f'{scheme}://{auth}{host}/{dbname}'



import copy


def passthrough(cls):
    """ Wrap another dictionary-like class to provide direct access to
    its contents through attribute access. The outer class is renamed
    to match the wrapped class, and the wrapped class is renamed to
    "Meta".
    """
    class Wrapper:
        def __init__(self, *args, **kwargs):
            self.__dict__['Meta'] = self.Meta(*args, **kwargs)

        def __getattr__(self, key):
            return self.Meta.get(key)

        def __setattr__(self, key, value):
            return self.Meta.set(key, value)

        def __getitem__(self, key):
            return self.Meta.get(key)

        def __setitem__(self, key, value):
            return self.Meta.set(key, value)

        def __deepcopy__(self, memo):
            # TODO: This process does one Meta construct too many.
            dup = type(self)()
            dup.__dict__['Meta'] = copy.deepcopy(self.Meta, memo)
            return dup

    Wrapper.__qualname__ = Wrapper.__name__ = cls.__name__
    cls.__qualname__ = cls.__name__ = 'Meta'
    Wrapper.Meta = cls
    return Wrapper


class PassthroughMetaclass:
    def __new__(meta, name, bases, attrs):
        cls = super().__new__(meta, name, bases, attrs)
        return passthrough(cls)



from polecat.cli.admin import *  # noqa
from polecat.cli.app import *  # noqa
from polecat.cli.aws import *  # noqa
from polecat.cli.build import *  # noqa
from polecat.cli.certificate import *  # noqa
from polecat.cli.db import *  # noqa
from polecat.cli.deploy import *  # noqa
from polecat.cli.deployment import *  # noqa
from polecat.cli.example import *  # noqa
# from polecat.cli.graphql import *  # noqa
from polecat.cli.main import main
from polecat.cli.migrate import *  # noqa
from polecat.cli.project import *  # noqa
from polecat.cli.publish import *  # noqa
from polecat.cli.secret import *  # noqa
from polecat.cli.server import *  # noqa
from polecat.cli.start import *  # noqa
from polecat.cli.test import *  # noqa
from polecat.cli.upload import *  # noqa


def entrypoint():
    main(auto_envvar_prefix='POLECAT', obj={})


if __name__ == '__main__':
    entrypoint()



import os

from botocore.exceptions import ClientError

from ...utils import set_path, substitute
from .exceptions import EntityDoesNotExist
from .utils import (aws_client, aws_does_not_exist, aws_exists, aws_resource,
                    default_path)


def create_policy(name, document, path=None, iam=None):
    iam = aws_client('iam', iam)
    with aws_exists():
        iam.create_policy(
            PolicyName=name,
            Path=default_path(path),
            PolicyDocument=document
        )


def delete_policy(name, iam=None):
    iam = aws_client('iam', iam)
    with aws_does_not_exist():
        arn = find_policy(name).arn
        iam.delete_policy(PolicyArn=arn)


def create_group(name, path=None, iam=None):
    iam = aws_client('iam', iam)
    with aws_exists():
        iam.create_group(
            GroupName=name,
            Path=default_path(path)
        )


def delete_group(name, iam=None):
    iam = aws_client('iam', iam)
    with aws_does_not_exist():
        iam.delete_group(GroupName=name)


def attach_group_policy(group_name, policy_arn, iam=None):
    iam = aws_client('iam', iam)
    with aws_exists():
        iam.attach_group_policy(
            GroupName=group_name,
            PolicyArn=policy_arn
        )


def detach_group_policy(group_name, policy_arn, iam=None):
    iam = aws_client('iam', iam)
    with aws_does_not_exist():
        iam.detach_group_policy(
            GroupName=group_name,
            PolicyArn=policy_arn
        )


def find_policy(name, iam=None):
    iam = aws_resource('iam', iam)
    for policy in iam.policies.filter():
        if policy.policy_name == name:
            return policy
    raise EntityDoesNotExist


def create_user(name, path=None, iam=None):
    iam = aws_client('iam', iam)
    with aws_exists():
        iam.create_user(
            UserName=name,
            Path=default_path(path)
        )
        response = iam.create_access_key(
            UserName=name
        )
        return response['AccessKey']


def add_user_to_group(user_name, group_name, iam=None):
    iam = aws_client('iam', iam)
    with aws_exists():
        iam.add_user_to_group(
            UserName=user_name,
            GroupName=group_name
        )


def create_bucket(name, s3=None):
    s3 = aws_client('s3', s3)
    with aws_exists():
        s3.create_bucket(
            Bucket=name,
            CreateBucketConfiguration={
                'LocationConstraint': os.environ['AWS_DEFAULT_REGION']
            }
        )


def attach_bucket_policy(bucket, policy, s3=None):
    s3 = aws_client('s3', s3)
    with aws_exists():
        s3.put_bucket_policy(
            Bucket=bucket,
            Policy=substitute(policy, bucket=bucket)
        )


def set_parameter(key, value, ssm=None):
    ssm = aws_client('ssm', ssm)
    if isinstance(value, (list, tuple)):
        type = 'StringList'
        value = ','.join(value)
    else:
        type = 'String'
    with aws_exists():
        ssm.put_parameter(
            Name=key,
            Type=type,
            Value=value,
            Overwrite=True
        )


def get_parameter(key, default=None, ssm=None):
    ssm = aws_client('ssm', ssm)
    try:
        response = ssm.get_parameter(Name=key)
    except ClientError as e:
        if e.response['Error']['Code'] == 'ParameterNotFound':
            return default
        raise
    param = response['Parameter']
    value = param['Value']
    if param['Type'] == 'StringList':
        value = value.split(',')
    return value


def get_parameters_by_path(path, ssm=None):
    if path[-1] != '/':
        raise ValueError('path must end in /')
    ssm = aws_client('ssm', ssm)
    try:
        response = ssm.get_parameters_by_path(Path=path, Recursive=True)
    except ClientError as e:
        if e.response['Error']['Code'] == 'ParameterNotFound':
            raise EntityDoesNotExist
        raise
    params = {}
    for param in response['Parameters']:
        set_path(params, param['Name'][len(path):], param['Value'], '/')
    return params


def delete_parameter(key, ssm=None):
    ssm = aws_client('ssm', ssm)
    try:
        if isinstance(key, (list, tuple)):
            ssm.delete_parameters(Names=key)
        else:
            ssm.delete_parameter(Name=key)
    except ClientError as e:
        if e.response['Error']['Code'] != 'ParameterNotFound':
            raise



from graphql import graphql_sync


def execute_query(schema, query, variables=None, reraise=False, context=None):
    result = graphql_sync(schema, query, variable_values=variables, context_value=context)
    if reraise and result.errors:
        raise result.errors[0]
    return result



from stdb.core.engine import default_engine as engine
from stdb.exceptions import StdbError
from stdb.utils import make_handle, strip_default

from .detail import Detail
from .query import Query


class Create(Query):
    valid_keys = set(['id', 'type', 'attributes', 'relationships'])
    relationship_types = set(['foreignkey', 'manytomany'])

    def __init__(self, version, endpoint, data=None, **kwargs):
        super().__init__(version, endpoint, **kwargs)
        self.data = data
        self._m2ms = []
        self._values = []
        self._replace_m2ms = False
        self._files = []
        self._update = False

    def execute(self):
        sql = self.get_sql()
        if not sql:
            return []
        sql = sql % tuple([
            engine.quote_literal(v)
            for v in self.get_args()
        ])

        try:
            r = engine.execute(sql, 1)
        except engine.plpy.spiexceptions.ForeignKeyViolation:
            raise StdbError('foreign key violation')

        # If any of our fields were files, run an action to confirm
        # each of them has a value in tmpfiles.
        if self._files:
            self.check_files(self._files)

        qry = Detail(self.version, self.endpoint, id=r[0]['id'])
        for r in qry.iter_rows():
            return r[1]

        # Arriving here means we have permission to create but not to
        # get details. Return emptyness.
        return {}

    def get_sql(self):
        sql = self.get_main_sql()
        m2m_sql = self.get_m2m_sql()
        if m2m_sql:
            if sql:
                m2m_parts = ', '.join(f't{i} AS ({m})' for i, m in enumerate(m2m_sql))
                sql = f'WITH inserted AS ({sql}), {m2m_parts} SELECT id FROM inserted;'
            else:
                sql = '; '.join(m2m_sql)
        return sql

    def get_main_sql(self):
        # TODO: Maybe optimise this lookup?
        table = self.endpoint['table']
        field_names, values = self.get_fields_values()
        if not field_names or not values:
            sql = f'INSERT INTO {table} (id) VALUES(DEFAULT) RETURNING id'
        else:
            f_str = ', '.join(field_names)
            v_str = ', '.join(['%s' for _ in values])
            sql = f'INSERT INTO {table} ({f_str}) VALUES({v_str}) RETURNING id'
        return sql

    def get_m2m_sql(self):
        sql = []
        for field_name, m2m_table, values in self._m2ms:
            if self._replace_m2ms:
                sql.append(f'DELETE FROM {m2m_table} USING inserted WHERE lid = inserted.id')
            if values:
                self._values.extend(values)
                v_str = ','.join(f'((SELECT id FROM inserted), %s)' for v in values)
                sql.append(f'INSERT INTO {m2m_table} (lid, rid) VALUES {v_str}')
        return sql

    def get_fields_values(self):
        field_names, values = [], []

        # Extract root-level data.
        try:
            root = self.data['data']
        except (TypeError, KeyError):
            raise StdbError('missing data')

        # Check for an ID first. This is okay as we want to be able
        # to support importing data that comes with IDs. However, we
        # need to ignore it if it's a UUID (TODO: add a link to the
        # relevant section of the JSON-API docs).
        if 'id' in root:
            try:
                v = int(root['id'])
                field_names.append('id')
                values.append(v)
            except ValueError:
                pass

        for key, data in root.items():
            if key == 'attributes':

                for name, value in data.items():

                    # Check for a known field.
                    try:
                        fld = self.endpoint['fields'][name]
                    except KeyError:
                        raise StdbError(f'unknown attribute: {name}')

                    # Must be an attribute.
                    if fld['field']['type'] in self.relationship_types:
                        raise StdbError(f'field is not an attribute: {name}')

                    # If the field is a file, add a post action to confirm the
                    # file exists in our temporary files cache.
                    if fld['field']['type'] == 'file':
                        self._files.append((name, value))

                    values.append(value)
                    field_names.append(fld['column'])

            elif key == 'relationships':

                for name, value in data.items():

                    # Check for a known field.
                    try:
                        fld = self.endpoint['fields'][name]
                    except KeyError:
                        raise StdbError(f'unknown relationship: {name}')

                    # Not allowed to set reverse fields here.
                    if fld['field']['type'] == 'reverse':
                        raise StdbError(f'cannot set reverse relationship: {name}')

                    if fld['field']['type'] == 'manytomany':

                        # Must have the endpoint other set.
                        # TODO: Fix duplicate below.
                        if not fld['other']:
                            raise StdbError('no value for "other" on endpoint')

                        self._m2ms.append((
                            name,
                            fld['m2m_table'],
                            self.parse_manytomany_value(fld, name, value)
                        ))
                    elif fld['field']['type'] == 'foreignkey':

                        # Must have the endpoint other set.
                        if not fld['other']:
                            raise StdbError('no value for "other" on endpoint')

                        values.append(self.parse_foreignkey_value(fld, name, value))
                        field_names.append(fld['column'])
                    else:
                        raise StdbError(f'field is not a relationship: {name}')

            elif key not in self.valid_keys:
                raise StdbError(f'unknown data key: {key}')

        self._values = values
        return field_names, values

    def get_args(self):
        return self._values

    def parse_foreignkey_value(self, field, name, value):
        try:
            v = value['data']
        except (TypeError, KeyError):
            raise StdbError((
                'invalid foreign-key value for "{}": missing "data"'
            ).format(
                name
            ))
        if v is None:
            return None
        try:
            type = make_handle(v['type'])
        except (TypeError, KeyError):
            raise StdbError((
                'invalid foreign-key value for "{}": missing "type"'
            ).format(
                name
            ))
        if type != field['other']:
            raise StdbError((
                'invalid foreign-key value for "{}": should be type "{}"'
            ).format(
                name, strip_default(field['other'])
            ))
        try:
            v = v['id']
        except (TypeError, KeyError):
            raise StdbError((
                'invalid foreign-key value for "{}": missing "id"'
            ).format(
                name
            ))
        return v

    def parse_manytomany_value(self, field, name, value):
        try:
            value = value['data']
        except (TypeError, KeyError):
            raise StdbError((
                'invalid many-to-many value for "{}": missing "data"'
            ).format(
                name
            ))
        if not isinstance(value, list):
            raise StdbError((
                'invalid many-to-many value for "{}": data must be a list'
            ).format(
                name
            ))
        vals = []
        for v in value:
            try:
                type = make_handle(v['type'])
            except (TypeError, KeyError):
                raise StdbError((
                    'invalid many-to-many value for "{}": missing "type"'
                ).format(
                    name
                ))
            if type != field['other']:
                raise StdbError((
                    'invalid many-to-many value for "{}": should be type "{}"'
                ).format(
                    name, strip_default(field['other'])
                ))
            try:
                v = v['id']
            except (TypeError, KeyError):
                raise StdbError((
                    'invalid many-to-many value for "{}": missing "id"'
                ).format(
                    name
                ))
            vals.append(v)
        return vals

    def check_files(self, files):
        """ Confirm each entry in files exists in the "tmpfiles" table.

        Also deletes those entries from the table.
        """
        if self._update:
            raise StdbError('updating files not implemented yet')
        dp_id = self.endpoint['datapoint_id']
        where = ' OR '.join([
            '(field = {} AND path = {})'.format(
                engine.quote_literal(f),
                engine.quote_literal(p)
            )
            for f, p in files
        ])
        sql = (
            'DELETE FROM file.tmpfiles'
            ' WHERE datapoint = {}'
            ' AND ({})'
            ' RETURNING field'
        ).format(
            engine.quote_literal(dp_id),
            where,
            engine.quote_literal([f[1] for f in files])
        )
        rslts = engine.execute(sql, len(files))
        missing = set([f[0] for f in files]) - set([r['field'] for r in rslts])
        if len(missing):
            raise StdbError('invalid files given for: {}'.format(', '.join(missing)))



from psycopg2.sql import SQL, Identifier

from .expression import Expression
from .value import value_to_sql


class Insert(Expression):
    def __init__(self, relation, values=None, returning=None):
        self.relation = relation
        self.values = values
        self.returning = returning or ()
        if 'id' not in self.returning:
            self.returning += ('id',)

    @property
    def alias(self):
        return self.relation.alias

    def to_sql(self):
        # TODO: Generating SQL for inserts could be improved by using
        # a Values expression instead of this mishmash.
        if not self.values or (isinstance(self.values, Expression) and self.values.values.values == [[]]):
            prefix_sql = SQL('DEFAULT VALUES')
            suffix_sql = SQL('')
            get_values_func = None
        elif isinstance(self.values, Expression):
            prefix_sql = SQL('')
            suffix_sql = SQL('')
            get_values_func = self.get_values_sql_from_expression
        else:
            prefix_sql = SQL('VALUES (')
            suffix_sql = SQL(')')
            get_values_func = self.get_values_sql_from_dict
        if get_values_func:
            column_names_sql, column_values_sql, column_values = get_values_func()
            column_names_sql = SQL('(') + SQL(', ').join(column_names_sql) + SQL(') ')
        else:
            column_names_sql = SQL('')
            column_values_sql = ()
            column_values = ()
        returning_sql = map(Identifier, self.returning)
        # TODO: This is a bit gross.
        sql = SQL('INSERT INTO {} {}{}{}{} RETURNING {}').format(
            Identifier(self.relation.alias),
            column_names_sql,
            prefix_sql,
            SQL(', ').join(column_values_sql),
            suffix_sql,
            SQL(', ').join(returning_sql)
        )
        return sql, tuple(column_values)

    def get_values_sql_from_expression(self):
        column_names_sql = []
        for column_name in self.values.iter_column_names():
            column_names_sql.append(Identifier(column_name))
        column_values_sql, column_values = self.values.to_sql()
        return column_names_sql, [column_values_sql], column_values

    def get_values_sql_from_dict(self):
        column_names_sql = []
        column_values_sql = []
        column_values = []
        for column_name, column_value in self.values.items():
            column = self.relation.get_column(column_name)
            column_names_sql.append(Identifier(column_name))
            value_sql, value = value_to_sql(column, column_value)
            column_values_sql.append(value_sql)
            column_values.extend(value)
        return column_names_sql, column_values_sql, column_values

    def get_subrelation(self, name):
        # TODO: This feels like bad design. Not sure how to correct
        # just at the moment, and at least it's clear what it's doing.
        return self.relation.get_subrelation(name)

    def get_column(self, name):
        # TODO: Also bad design, but also what's the difference
        # between this and the above? Looks like the above just does
        # 'this.related_table'.
        # TODO: This should be filtered by "returning".
        return self.relation.get_column(name)

    def has_column(self, name):
        # TODO: Also bad design, but also what's the difference
        # between this and the above? Looks like the above just does
        # 'this.related_table'.
        # TODO: This should be filtered by "returning".
        return self.relation.has_column(name)

    def push_selection(self, selection=None):
        if selection:
            # TODO: Efficiency.
            for column_name in selection:
                if column_name not in self.returning:
                    self.returning += (column_name,)



from psycopg2.sql import SQL, Identifier

from .expression import Expression


class TrigramDistance(Expression):
    def __init__(self, column, term):
        self.column = column
        self.term = term

    def to_sql(self):
        return SQL('{} <-> %s').format(
            Identifier(self.column)
        ), (self.term,)



from termcolor import colored

from ...utils.feedback import feedback
from .constants import DEPLOYMENT_PREFIX, DEPLOYMENT_REGISTRY
from .exceptions import EntityDoesNotExist, EntityExists
from .operations import delete_parameter, get_parameter, set_parameter
from .project import assert_project_exists
from .utils import aws_client


@feedback
def create_deployment(project, deployment, feedback):
    ssm = aws_client('ssm')
    with feedback(f'Create deployment {colored(project, "cyan")}/{colored(deployment, "green")}'):
        assert_project_exists(project, ssm=ssm)
        name = DEPLOYMENT_REGISTRY.format(project)
        deployments = get_parameter(name, default=[], ssm=ssm)
        if deployment in deployments:
            raise EntityExists
        try:
            deployments.remove('_')
        except ValueError:
            pass
        deployments.append(deployment)
        set_parameter(name, deployments, ssm=ssm)
        # The active flag is here really only to ensure when
        # deployments are carried out for deployments without secrets
        # etc, they are still used.
        name = DEPLOYMENT_PREFIX.format(project, deployment) + 'active'
        set_parameter(name, '1', ssm=ssm)


@feedback
def list_deployments(project, feedback):
    ssm = aws_client('ssm')
    with feedback(f'List deployments for {colored(project, "cyan")}'):
        assert_project_exists(project, ssm=ssm)
        name = DEPLOYMENT_REGISTRY.format(project)
        deployments = get_parameter(name, default=[], ssm=ssm)
        return [d for d in deployments if d != '_']


@feedback
def delete_deployment(project, deployment, feedback):
    ssm = aws_client('ssm')
    with feedback(f'Delete deployment {colored(project, "cyan")}/{colored(deployment, "green")}'):
        assert_project_exists(project, ssm=ssm)
        assert_deployment_exists(project, deployment, ssm=ssm)
        # TODO: Check for related entities?
        name = DEPLOYMENT_REGISTRY.format(project)
        deployments = get_parameter(name, default=[], ssm=ssm)
        deployments.remove(deployment)
        if not deployments:
            deployments.append('_')
        set_parameter(name, deployments, ssm=ssm)
        # TODO: Need to fetch the path, then delete all of them.
        name = DEPLOYMENT_PREFIX.format(project, deployment)
        delete_parameter(name, ssm=ssm)


def deployment_exists(project, deployment, ssm=None):
    registry = get_parameter(DEPLOYMENT_REGISTRY.format(project), ssm=ssm)
    return deployment in registry


def assert_deployment_exists(project, deployment, ssm=None):
    if not deployment_exists(project, deployment, ssm=ssm):
        raise EntityDoesNotExist(
            f'deployment {deployment} does not exist'
        )



class Auto:
    pass


def table_to_identifier(table):
    if isinstance(table, str):
        ident = table
    else:
        ident = table.name
        # if table.app:
        #     ident = f'{table.app.name}.{ident}'
    return ident


def column_to_identifier(column):
    if isinstance(column, str):
        ident = column
    else:
        ident = column.name
    return ident



import os
from pathlib import Path

import click
import ujson


def load_config():
    cfg = read_config(global_config_path())
    path = find_config()
    cfg.update(read_config(path))
    return cfg, path


def update_config(new_cfg={}, delete=[], path=None):
    if path is None:
        path = find_config()
        if not path:
            path = os.path.join(os.getcwd(), 'polecat.json')
    cfg = read_config(path)
    for d in delete:
        try:
            del cfg[d]
        except KeyError:
            pass
    cfg.update(new_cfg)
    write_config(path, cfg)


def find_config():
    dir = Path().absolute()
    while True:
        path = dir / 'polecat.json'
        if path.exists():
            return str(path)
        if dir == Path.home() or dir.parent == dir:
            return None
        dir = dir.parent


def global_config_path():
    return os.path.join(click.get_app_dir('polecat'), 'polecat.json')


def read_config(path):
    cfg = {}
    if path and os.path.exists(path):
        data = open(path, 'r').read()
        if data:
            cfg = ujson.loads(data)
    return cfg


def write_config(path, cfg):
    if not os.path.exists(path):
        try:
            os.makedirs(os.path.dirname(path))
        except Exception:
            pass
        try:
            open(path, 'x')
        except Exception:
            pass
    os.chmod(path, 0o600)
    cfg_str = ujson.dumps(cfg, indent=2)
    open(path, 'w').write(cfg_str)



from functools import partial

from polecat.db.query import query as query_module
from psycopg2.sql import SQL, Placeholder

from .expression import Expression


class Values(Expression):
    def __init__(self, values, relation=None):
        self.values = values
        self.relation = relation
        self.keyword = 'VALUES'

    def to_sql(self):
        if isinstance(self.values, query_module.Values):
            get_values_sql = partial(
                self.get_values_sql_from_values, self.values
            )
        else:
            get_values_sql = partial(
                self.get_values_sql_from_dict, self.values
            )
        return self.get_values_sql(get_values_sql)

    def get_values_sql(self, get_values_sql):
        values_sql, values_args = get_values_sql()
        joined_sql = SQL(', ').join(
            SQL('({})').format(
                SQL(', ').join(row_sql)
            )
            for row_sql in values_sql
        )
        return SQL('%s {}' % self.keyword).format(joined_sql), values_args

    def get_values_sql_from_values(self, values):
        column_values_sql = []
        column_values = ()
        for row in values.iter_rows():
            row_values_sql = []
            for column_name, column_value in row:
                value_sql, value = self.value_to_sql(column_value, column_name)
                row_values_sql.append(value_sql)
                column_values += value
            column_values_sql.append(row_values_sql)
        return column_values_sql, column_values

    def get_values_sql_from_dict(self, values_dict):
        column_values_sql = []
        column_values = ()
        for column_name, column_value in values_dict.items():
            value_sql, value = self.value_to_sql(column_value, column_name)
            column_values_sql.append(value_sql)
            column_values += value
        return (column_values_sql,), column_values

    def value_to_sql(self, value, column_name=None):
        if isinstance(value, Expression):
            sql, args = value.to_sql()
            return SQL('{}').format(sql), args
        else:
            if self.relation and column_name:
                column = self.relation.get_column(column_name)
                value = column.to_db_value(value)
            return Placeholder(), (value,)

    def iter_column_names(self):
        if isinstance(self.values, dict):
            return self.values.keys()
        else:
            return self.values.iter_column_names()



from polecat.db.schema.variable import SessionVariable

__all__ = ('session',)


class Session:
    def __getitem__(self, key):
        return SessionVariable(key)

    def __call__(self, key, type):
        return SessionVariable(key, type)


session = Session()



from psycopg2.sql import SQL, Identifier

from .as_ import As
from .expression import Expression


class CTE(Expression):
    def __init__(self):
        self.expression = None
        self.common_expressions = []
        self.alias_counter = 0

    def append(self, expression):
        alias = self.create_alias(expression)
        self.common_expressions.append(alias)
        return alias

    def prepend(self, expression):
        alias = self.create_alias(expression)
        self.expressions.insert(0, expression)
        return alias

    def set_final_expression(self, expression):
        self.expression = expression

    def create_alias(self, expression):
        alias = CTEAs(expression, f'c{self.alias_counter}')
        self.alias_counter += 1
        return alias

    def to_sql(self):
        if self.common_expressions:
            return self.to_cte_sql()
        else:
            return self.expression.to_sql()

    def to_cte_sql(self):
        common_sql, common_args = self.get_common_sql()
        final_sql, final_args = self.expression.to_sql()
        return (
            SQL('WITH {} {}').format(
                SQL(', ').join(common_sql),
                final_sql
            ),
            common_args + final_args
        )

    def get_common_sql(self):
        common_sql = []
        common_args = ()
        for expr in self.common_expressions:
            sql, args = expr.to_sql()
            common_sql.append(sql)
            common_args += args
        return common_sql, common_args

    def iter_expressions(self):
        for expr in self.common_expressions:
            yield expr
        yield self.expression


class CTEAs(As):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.recursive_expression = None

    def get_as_sql(self, expression_sql):
        rec_sql, rec_args = self.get_recursive_sql()
        return SQL('{}{} AS ({}{})').format(
            SQL('RECURSIVE ' if self.recursive_expression else ''),
            Identifier(self.alias),
            expression_sql,
            rec_sql
        ), rec_args

    def get_recursive_sql(self):
        if self.recursive_expression:
            expr_sql, expr_args = self.recursive_expression.to_sql()
            return SQL(' UNION ALL {}').format(
                expr_sql
            ), expr_args
        else:
            return SQL(''), ()

    def set_recursive_expression(self, expression):
        self.recursive_expression = expression



import inspect
from pathlib import Path

from polecat.core.registry import Registry, RegistryMetaclass

Registry('app_registry', construct=True)


class App(metaclass=RegistryMetaclass):
    _registry = 'app_registry'
    _registry_base = 'App'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = self.__class__.__name__
        if self.name.endswith('App'):
            self.name = self.name[:-3]
        self.models = []
        self.roles = []
        self.types = []
        self.access = []

    def __eq__(self, other):
        if not isinstance(other, App):
            return False
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    @property
    def path(self):
        return Path(inspect.getfile(self.__class__)).parent



from termcolor import colored

from ...utils.feedback import feedback
from .constants import CERTIFICATE, ZONE
from .operations import delete_parameter, set_parameter
from .utils import aws_client


@feedback
def publish(project, deployment, domain, certificate, zone, feedback):
    if not certificate:
        certificate = domain
    if not zone:
        # TODO: What is this thing actually doing?
        # zone = domain[:domain.rindex('.', domain.rindex('.') - 1)]
        zone = domain + '.'
    acm = aws_client('acm', region='us-east-1')
    r53 = aws_client('route53')
    ssm = aws_client('ssm')
    cert_arn = None
    possible_certs = []
    zone_id = None
    with feedback(f'Find certificate matching {colored(certificate, "yellow")}'):
        response = acm.list_certificates()
        for cert in response['CertificateSummaryList']:
            tags = acm.list_tags_for_certificate(
                CertificateArn=cert['CertificateArn']
            )['Tags']
            name_tag = [
                t for t in tags
                if t['Key'] == 'Name' and t['Value'] == certificate
            ]
            if len(name_tag) > 0:
                cert_arn = cert['CertificateArn']
                break
            elif cert['DomainName'] == certificate:
                cert_arn = cert['CertificateArn']
                break
            possible_certs.append(cert['DomainName'])
        if not cert_arn:
            raise Warning('failed')
    with feedback(f'Find hosted zone matching {colored(zone, "blue")}'):
        response = r53.list_hosted_zones()
        for hz in response['HostedZones']:
            if hz['Name'] == zone:
                zone_id = hz['Id']
        if not zone_id:
            raise Warning('failed')
    if not cert_arn:
        print(possible_certs)
        return
    with feedback(f'Publish {colored(project, "blue")}/{colored(deployment, "green")} to {colored(domain, "yellow")}'):
        name = CERTIFICATE.format(project, deployment, domain)
        set_parameter(name, cert_arn, ssm=ssm)
        name = ZONE.format(project, deployment, domain)
        if zone_id:
            set_parameter(name, str(zone_id), ssm=ssm)
        else:
            delete_parameter(name, ssm=ssm)


@feedback
def unpublish(project, deployment, domain, feedback):
    # TODO
    pass



from ...utils import capitalize
from .constants import MEDIA_PREFIX, SERVER_CODE


def create_api_resources(project, deployment, api_name, bucket, environment):
    project_deployment = f'{capitalize(project)}{capitalize(deployment)}'
    code_version = environment['code']['version']
    bundle_version = environment['bundle']['version']
    secrets = environment.get('secrets', {})
    return {
        f'{project_deployment}ServerLambda': {
            'Type': 'AWS::Lambda::Function',
            'Properties': {
                'Code': {
                    'S3Bucket': bucket,
                    'S3Key': SERVER_CODE.format(project, code_version)
                },
                'Description': f'Polecat API resolver for {project}/{deployment}',
                'FunctionName': f'{project_deployment}Server',
                'Handler': f'main.handler',
                'Timeout': 30,
                'Role': {
                    'Fn::GetAtt': [
                        f'{project_deployment}ExecutionRole',
                        'Arn'
                    ]
                },
                'Runtime': 'python3.7',
                'Environment': {
                    'Variables': {
                        'PROJECT': project,
                        'DEPLOYMENT': deployment,
                        'BUCKET': bucket,
                        'MEDIA_PREFIX': MEDIA_PREFIX.format(
                            project,
                            deployment
                        ),
                        'CODE_VERSION': code_version,
                        'BUNDLE_VERSION': bundle_version,
                        **secrets
                    }
                },
                'Tags': [
                    {
                        'Key': 'Builder',
                        'Value': 'Polecat'
                    },
                    {
                        'Key': 'PolecatProject',
                        'Value': project
                    },
                    {
                        'Key': 'PolecatDeployment',
                        'Value': deployment
                    }
                ]
            }
        },
        f'{project_deployment}ExecutionRole': {
            'Type': 'AWS::IAM::Role',
            'Properties': {
                'AssumeRolePolicyDocument': {
                    'Version': '2012-10-17',
                    'Statement': [
                        {
                            'Effect': 'Allow',
                            'Principal': {
                                'Service': [
                                    'lambda.amazonaws.com'
                                ]
                            },
                            'Action': [
                                'sts:AssumeRole'
                            ]
                        }
                    ]
                },
                'ManagedPolicyArns': [
                    'arn:aws:iam::aws:policy/service-role/AWSLambdaBasicExecutionRole'
                ],
                'Policies': [
                    {
                        'PolicyName': f'{project_deployment}MediaAccess',
                        'PolicyDocument': {
                            'Version': '2012-10-17',
                            'Statement': [
                                {
                                    'Effect': 'Allow',
                                    'Action': [
                                        's3:DeleteObject',
                                        's3:DeleteObjectVersion',
                                        's3:GetObject',
                                        's3:PutObject'
                                    ],
                                    'Resource': {
                                        'Fn::Join': [
                                            '',
                                            [
                                                # TODO: Use a constant.
                                                f'arn:aws:s3:::{bucket}/' + MEDIA_PREFIX.format(project, deployment) + '*'
                                            ]
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                ]
            }
        },
        f'{api_name}Api': {
            'Type': 'AWS::ApiGateway::RestApi',
            'Properties': {
                'Name': api_name,
                'Description': f'Polecat API Gateway',
                'FailOnWarnings': True,
                'EndpointConfiguration': {
                    'Types': [
                        # TODO: This was regional before... problem?
                        'EDGE'
                    ]
                }
            }
        },
        f'{project_deployment}ServerLambdaPermission': {
            'Type': 'AWS::Lambda::Permission',
            'Properties': {
                'Action': 'lambda:invokeFunction',
                'FunctionName': {
                    'Fn::GetAtt': [
                        f'{project_deployment}ServerLambda',
                        'Arn'
                    ]
                },
                'Principal': 'apigateway.amazonaws.com',
                'SourceArn': {
                    'Fn::Join': [
                        '',
                        [
                            'arn:aws:execute-api:',
                            {
                                'Ref': 'AWS::Region'
                            },
                            ':',
                            {
                                'Ref': 'AWS::AccountId'
                            },
                            ':',
                            {
                                'Ref': f'{api_name}Api'
                            },
                            '/*'
                        ]
                    ]
                }
            }
        },
        f'{capitalize(project)}ApiGatewayCloudWatchLogsRole': {
            'Type': 'AWS::IAM::Role',
            'Properties': {
                'AssumeRolePolicyDocument': {
                    'Version': '2012-10-17',
                    'Statement': [
                        {
                            'Effect': 'Allow',
                            'Principal': {
                                'Service': [
                                    'apigateway.amazonaws.com'
                                ]
                            },
                            'Action': [
                                'sts:AssumeRole'
                            ]
                        }
                    ]
                },
                'Policies': [
                    {
                        'PolicyName': 'ApiGatewayLogsPolicy',
                        'PolicyDocument': {
                            'Version': '2012-10-17',
                            'Statement': [
                                {
                                    'Effect': 'Allow',
                                    'Action': [
                                        'logs:CreateLogGroup',
                                        'logs:CreateLogStream',
                                        'logs:DescribeLogGroups',
                                        'logs:DescribeLogStreams',
                                        'logs:PutLogEvents',
                                        'logs:GetLogEvents',
                                        'logs:FilterLogEvents'
                                    ],
                                    'Resource': '*'
                                }
                            ]
                        }
                    }
                ]
            }
        },
        f'{project_deployment}ApiGatewayAccount': {
            'Type': 'AWS::ApiGateway::Account',
            'Properties': {
                'CloudWatchRoleArn': {
                    'Fn::GetAtt': [
                        f'{capitalize(project)}ApiGatewayCloudWatchLogsRole',
                        'Arn'
                    ]
                }
            }
        },
        f'{project_deployment}ApiStage': {
            'DependsOn': [
                f'{project_deployment}ApiGatewayAccount'
            ],
            'Type': 'AWS::ApiGateway::Stage',
            'Properties': {
                'DeploymentId': {
                    'Ref': f'{project_deployment}ApiDeployment'
                },
                'MethodSettings': [
                    {
                        'DataTraceEnabled': True,
                        'HttpMethod': '*',
                        'LoggingLevel': 'INFO',
                        'ResourcePath': '/*'
                    }
                ],
                'RestApiId': {
                    'Ref': f'{api_name}Api'
                },
                'StageName': f'{project}-{deployment}',
                'Tags': [
                    {
                        'Key': 'Builder',
                        'Value': 'Polecat'
                    },
                    {
                        'Key': 'PolecatProject',
                        'Value': project
                    },
                    {
                        'Key': 'PolecatDeployment',
                        'Value': deployment
                    }
                ]
            }
        },
        f'{project_deployment}ApiDeployment': {
            'Type': 'AWS::ApiGateway::Deployment',
            'DependsOn': [
                f'{project_deployment}Request'
            ],
            'Properties': {
                'RestApiId': {
                    'Ref': f'{api_name}Api'
                }
            }
        },
        f'{project_deployment}Resource': {
            'Type': 'AWS::ApiGateway::Resource',
            'Properties': {
                'RestApiId': {
                    'Ref': f'{api_name}Api'
                },
                'ParentId': {
                    'Fn::GetAtt': [
                        f'{api_name}Api', 'RootResourceId'
                    ]
                },
                'PathPart': '{proxy+}'
            }
        },
        f'{project_deployment}Request': {
            'DependsOn': f'{project_deployment}ServerLambdaPermission',
            'Type': 'AWS::ApiGateway::Method',
            'Properties': {
                'ApiKeyRequired': False,
                'AuthorizationType': 'NONE',
                'HttpMethod': 'ANY',
                'Integration': {
                    'Type': 'AWS_PROXY',
                    'IntegrationHttpMethod': 'POST',
                    'Uri': {
                        'Fn::Join': [
                            '',
                            [
                                'arn:aws:apigateway:',
                                {
                                    'Ref': 'AWS::Region'
                                },
                                ':lambda:path/2015-03-31/functions/',
                                {
                                    'Fn::GetAtt': [
                                        f'{project_deployment}ServerLambda',
                                        'Arn'
                                    ]
                                },
                                '/invocations'
                            ]
                        ]
                    },
                    'IntegrationResponses': [
                        {
                            'StatusCode': 200
                        }
                    ]
                },
                'ResourceId': {
                    'Ref': f'{project_deployment}Resource'
                },
                'RestApiId': {
                    'Ref': f'{api_name}Api'
                },
                'MethodResponses': [
                    {
                        'StatusCode': 200
                    }
                ]
            }
        },
        f'{project_deployment}RootRequest': {
            'DependsOn': f'{project_deployment}ServerLambdaPermission',
            'Type': 'AWS::ApiGateway::Method',
            'Properties': {
                'ApiKeyRequired': False,
                'AuthorizationType': 'NONE',
                'HttpMethod': 'ANY',
                'Integration': {
                    'Type': 'AWS_PROXY',
                    'IntegrationHttpMethod': 'POST',
                    'Uri': {
                        'Fn::Join': [
                            '',
                            [
                                'arn:aws:apigateway:',
                                {
                                    'Ref': 'AWS::Region'
                                },
                                ':lambda:path/2015-03-31/functions/',
                                {
                                    'Fn::GetAtt': [
                                        f'{project_deployment}ServerLambda',
                                        'Arn'
                                    ]
                                },
                                '/invocations'
                            ]
                        ]
                    },
                    'IntegrationResponses': [
                        {
                            'StatusCode': 200
                        }
                    ]
                },
                'ResourceId': {
                    'Fn::GetAtt': [
                        f'{api_name}Api',
                        'RootResourceId'
                    ]
                },
                'RestApiId': {
                    'Ref': f'{api_name}Api'
                },
                'MethodResponses': [
                    {
                        'StatusCode': 200
                    }
                ]
            }
        }
    }


def create_domain_resources(project, deployment, api_name, domain, certificate_arn):
    project_deployment = f'{capitalize(project)}{capitalize(deployment)}'
    return {
        f'{project_deployment}DomainName': {
            'DependsOn': f'{project_deployment}ApiDeployment',
            'Type': 'AWS::ApiGateway::DomainName',
            'Properties': {
                'DomainName': domain,
                'EndpointConfiguration': {
                    'Types': [
                        'EDGE'
                    ]
                },
                'CertificateArn': certificate_arn
            }
        },
        f'{project_deployment}BasePath': {
            'DependsOn': f'{project_deployment}DomainName',
            'Type': 'AWS::ApiGateway::BasePathMapping',
            'Properties': {
                'DomainName': domain,
                'RestApiId': {
                    'Ref': f'{api_name}Api'
                },
                'Stage': f'{project}-{deployment}'
            }
        }
    }


def create_zone_resources(project, deployment, domain, zone):
    project_deployment = f'{capitalize(project)}{capitalize(deployment)}'
    return {
        f'{project_deployment}RecordSet': {
            'DependsOn': f'{project_deployment}DomainName',
            'Type': 'AWS::Route53::RecordSet',
            'Properties': {
                'Name': domain,
                'Type': 'A',
                'HostedZoneId': zone,
                'AliasTarget': {
                    'DNSName': {
                        'Fn::GetAtt': [
                            f'{project_deployment}DomainName',
                            'DistributionDomainName'
                        ]
                    },
                    'HostedZoneId': {
                        'Fn::GetAtt': [
                            f'{project_deployment}DomainName',
                            'DistributionHostedZoneId'
                        ]
                    },
                    'EvaluateTargetHealth': False
                }
            }
        }
    }


def create_output_resources(project, deployment, api_name):
    project_deployment = f'{capitalize(project)}{capitalize(deployment)}'
    return {
        f'{project_deployment}Url': {
            'Description': f'Polecat root URL of {project}/{deployment} API gateway',
            'Value': {
                'Fn::Join': [
                    '',
                    [
                        'https://',
                        {
                            'Ref': f'{api_name}Api'
                        },
                        '.execute-api.',
                        {
                            'Ref': 'AWS::Region'
                        },
                        f'.amazonaws.com/{project}-{deployment}/'
                    ]
                ]
            }
        }
    }



import inspect
import logging
from functools import partial
from itertools import chain

from cached_property import cached_property

from polecat.utils import to_list

from .defaults import default_blueprint
from .db.query import Q

logger = logging.getLogger(__name__)


class APIContext:
    def __init__(self, event=None):
        self.event = event
        self.stack = []

    def parse_argument(self, name):
        raise NotImplementedError

    def parse_input(self):
        raise NotImplementedError

    @cached_property
    def model_class(self):
        return self.get_model_class()

    def get_model_class(self):
        raise NotImplementedError

    @cached_property
    def selector(self):
        return self.get_selector()

    def get_selector(self):
        raise NotImplementedError

    def get_model(self, name):
        # TODO: This is no good. It should include the app name.
        return default_blueprint.models[name]


class ResolverContext:
    def __init__(self, manager, api_context):
        self.manager = manager
        self.api_context = api_context
        self.stack = []

    def __call__(self, *args, **kwargs):
        logger.debug('ResolveContext: moving to next operation')
        return next(self)(*args, **kwargs)

    def __next__(self):
        while len(self.stack):
            try:
                return next(self.stack[-1])
            except StopIteration:
                self.stack.pop()
                if not len(self.stack):
                    raise

    def __getattr__(self, key):
        return getattr(self.api_context, key)

    def push_iterator(self, iterator):
        self.stack.append(iterator)

    def cut_point(self, *args, **kwargs):
        return self.manager.cut_point(*args, **kwargs)


class ResolverList:
    def __init__(self, resolvers=None):
        self.resolvers = []
        self.use(resolvers)

    def __iter__(self):
        return iter(self.resolvers)

    def use(self, resolvers):
        # TODO: Track origin of resolvers.
        resolvers = to_list(resolvers)
        self.resolvers = resolvers + self.resolvers


class ResolverManager:
    def __call__(self, api_context, *args, **kwargs):
        return self.resolve(api_context, *args, **kwargs)

    def resolve(self, api_context, *args, **kwargs):
        logger.debug(f'Resolve: {self.__class__.__name__}')
        context = ResolverContext(self, api_context)
        return self.resolve_with_context(context, *args, **kwargs)

    def resolve_with_context(self, context, *args, **kwargs):
        iterator = iter(ResolverIterator(context))
        return next(iterator)(*args, **kwargs)

    def cut_point(self, method_name, context, *args, **kwargs):
        iterator = iter(ResolverMethodIterator(context, method_name))
        return next(iterator)(*args, **kwargs)

    def iter_resolvers(self, context):
        raise NotImplementedError


class AllResolverManager(ResolverManager):
    def iter_resolvers(self, context):
        model_class = context.model_class
        return chain(
            model_class.Meta.query_resolvers,
            model_class.Meta.all_resolvers
        )


class GetResolverManager(ResolverManager):
    def iter_resolvers(self, context):
        model_class = context.model_class
        return chain(
            model_class.Meta.query_resolvers,
            model_class.Meta.get_resolvers
        )


class CreateResolverManager(ResolverManager):
    def iter_resolvers(self, context):
        model_class = context.model_class
        return chain(
            model_class.Meta.mutation_resolvers,
            model_class.Meta.create_resolvers
        )


class UpdateResolverManager(ResolverManager):
    def iter_resolvers(self, context):
        model_class = context.model_class
        return chain(
            model_class.Meta.mutation_resolvers,
            model_class.Meta.update_resolvers
        )


class UpdateOrCreateResolverManager(ResolverManager):
    def resolve_with_context(self, context, *args, **kwargs):
        model_class = context.model_class
        id = context.parse_argument('id')
        if id is None:
            manager = model_class.Meta.create_resolver_manager
        else:
            manager = model_class.Meta.update_resolver_manager
        # TODO: Not too sure about this hard override of manager.
        context.manager = manager
        return manager.resolve_with_context(context, *args, **kwargs)


class DeleteResolverManager(ResolverManager):
    def iter_resolvers(self, context):
        model_class = context.model_class
        return chain(
            model_class.Meta.delete_resolvers
        )


class QueryResolverManager(ResolverManager):
    def iter_resolvers(self, context):
        query = context.query
        # TODO: This is interesting. I'm not sure if I want to have to
        # instantiate the mutation here, or at all. It's nicer for the
        # user though.
        return chain(
            query.Meta.resolvers,
            (query().resolve,)
        )


class MutationResolverManager(ResolverManager):
    def iter_resolvers(self, context):
        mutation = context.mutation
        # TODO: This is interesting. I'm not sure if I want to have to
        # instantiate the mutation here, or at all. It's nicer for the
        # user though.
        return chain(
            mutation.resolvers,
            (mutation().resolve,)
        )


class ResolverIterator:
    def __init__(self, context):
        self.context = context
        self.context.push_iterator(self)

    def __iter__(self):
        logger.debug('ResolverIterator: initiating')
        self.resolver_iterator = self.context.manager.iter_resolvers(self.context)
        return self

    def __next__(self):
        logger.debug('ResolverIterator: moving to next resolver')
        while True:
            resolver = next(self.resolver_iterator)
            logger.debug(f'ResolverIterator: considering {resolver}')
            if hasattr(resolver, 'resolve'):
                logger.debug(f'ResolverIterator: accepting (has resolve method)')
                func = resolver.resolve
            elif not inspect.isclass(resolver) and callable(resolver):
                logger.debug(f'ResolverIterator: accepting (is callable)')
                func = resolver
            else:
                logger.debug(f'ResolverIterator: rejecting')
                continue
            return partial(func, self.context)


class ResolverMethodIterator(ResolverIterator):
    def __init__(self, context, method_name):
        super().__init__(context)
        self.method_name = method_name

    def __next__(self):
        while True:
            resolver = next(self.resolver_iterator)
            if not hasattr(resolver, self.method_name):
                continue
            return partial(getattr(resolver, self.method_name), self.context)


class QueryResolver:
    def resolve(self, context, query=None):
        query = context.cut_point('build_query', context, query)
        results = context.cut_point('build_results', context, query)
        return results

    def build_query(self, context, query=None):
        if not query:
            query = Q(context.model_class, session=context.session)
        return query

    def resolve_model_fields(self, context, model):
        model_class = context.model_class
        for name, field in model_class.Meta.fields.items():
            for resolver in field.query_resolvers:
                resolver(context, model=model, field=field, field_name=name)
        return model


class AllResolver(QueryResolver):
    def build_query(self, context, query=None):
        query = super().build_query(context, query)
        return query.select(context.selector)

    def build_results(self, context, query):
        return [
            self.resolve_model_fields(context, m)
            for m in query
        ]


class GetResolver(QueryResolver):
    def build_query(self, context, query=None):
        query = super().build_query(context, query)
        id = context.parse_argument('id')
        return (
            query
            .filter(id=id)
            .select(context.selector)
        )

    def build_results(self, context, query):
        return self.resolve_model_fields(context, query.get())


class MutationResolver:
    def resolve(self, context):
        model = context.cut_point('build_model', context)
        query = context.cut_point('build_query', context, model)
        query = context.cut_point('select_query', context, query)
        results = context.cut_point('build_results', context, query)
        return results

    def resolve_model_fields(self, context, model):
        for name, field in model.Meta.fields.items():
            for resolver in field.mutation_resolvers:
                resolver(context, model=model, field=field, field_name=name)

    def select_query(self, context, query):
        return (
            query
            .select(context.selector)
        )

    def build_results(self, context, query):
        return query.get()


class CreateResolver(MutationResolver):
    def build_model(self, context):
        model_class = context.model_class
        input = context.parse_input()
        model = model_class(**input)
        self.resolve_model_fields(context, model)
        return model

    def build_query(self, context, model):
        return Q(model, session=context.session).insert()


class UpdateResolver(CreateResolver):
    def build_model(self, context):
        model_class = context.model_class
        input = context.parse_input()
        context._id = context.parse_argument('id')
        model = model_class(id=context._id, **input)
        self.resolve_model_fields(context, model)
        return model

    def build_query(self, context, model):
        return Q(model, session=context.session).update()


class DeleteResolver:
    def resolve(self, context):
        model = context.cut_point('build_model', context)
        query = context.cut_point('build_query', context, model)
        results = context.cut_point('build_results', context, query)
        return results

    def build_model(self, context):
        model_class = context.model_class
        # TODO: Should be from arguments. :(
        context._id = context.raw_input()['id']
        return model_class(id=context._id)

    def build_query(self, context, model):
        return Q(model, session=context.session).delete()

    def build_results(self, context, query):
        query.execute()
        return {'id': context._id}



from functools import singledispatch

from ...db.schema import (BoolColumn, FloatColumn, IntColumn, JSONColumn,
                          PasswordColumn, PointColumn, RelatedColumn,
                          SerialColumn, TextColumn, TimestampColumn,
                          UUIDColumn, DateColumn)
from .. import field


def create_column_from_field(field, type, **kwargs):
    return type(
        field.name,
        unique=field.unique,
        null=field.null,
        primary_key=field.primary_key,
        default=field.default,
        **kwargs
    )


@singledispatch
def convert_field(field, builder):
    raise NotImplementedError


@convert_field.register(field.TextField)
@convert_field.register(field.EmailField)
@convert_field.register(field.PhoneField)
def convert_textfield(field, builder):
    return create_column_from_field(field, TextColumn, max_length=field.length)


@convert_field.register(field.PasswordField)
def convert_passwordfield(field, builder):
    return create_column_from_field(field, PasswordColumn)


@convert_field.register(field.BoolField)
def convert_boolfield(field, builder):
    return create_column_from_field(field, BoolColumn)


@convert_field.register(field.IntField)
def convert_intfield(field, builder):
    return create_column_from_field(field, IntColumn)


@convert_field.register(field.SerialField)
def convert_serialfield(field, builder):
    return create_column_from_field(field, SerialColumn)


@convert_field.register(field.FloatField)
def convert_floatfield(field, builder):
    return create_column_from_field(field, FloatColumn)


@convert_field.register(field.DateField)
def convert_datefield(field, builder):
    return create_column_from_field(field, DateColumn)


@convert_field.register(field.DatetimeField)
def convert_datetimefield(field, builder):
    return create_column_from_field(field, TimestampColumn)


@convert_field.register(field.UUIDField)
def convert_uuidfield(field, builder):
    return create_column_from_field(field, UUIDColumn)


@convert_field.register(field.JSONField)
def convert_jsonfield(field, builder):
    return create_column_from_field(field, JSONColumn)


@convert_field.register(field.PointField)
@convert_field.register(field.GCSPointField)
def convert_pointfield(field, builder):
    return create_column_from_field(field, PointColumn)


@convert_field.register(field.RelatedField)
def convert_relatedfield(field, builder):
    return RelatedColumn(
        field.name,
        builder.build(field.other),
        related_column=field.related_name,
        unique=field.unique,
        null=field.null,
        primary_key=field.primary_key,
        default=field.default,
        on_delete=field.on_delete
    )



from functools import wraps

from .exceptions import EntityDoesNotExist, EntityExists, KnownError


class Feedback:
    def __call__(self, message=None):
        self.update_message(message)
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type:
            if exc_type == EntityExists:
                self.exists(exc_value)
                return True
            elif exc_type == EntityDoesNotExist:
                self.errored(exc_value)
                return True
            elif exc_type == Warning:
                self.warning(exc_value)
                return True
            elif exc_type == KnownError:
                self.errored(exc_value)
                return True
        else:
            self.succeeded()

    def update_message(self, message):
        self.message = message or ''

    def errored(self, error):
        pass

    def succeeded(self):
        pass

    def warning(self, warning):
        pass

    def info(self, message):
        pass

    def exists(self, error):
        pass


def feedback(func):
    @wraps(func)
    def inner(*args, feedback=None, **kwargs):
        if feedback is None:
            fb = Feedback()
            return func(*args, feedback=fb, **kwargs)
        else:
            return func(*args, feedback=feedback, **kwargs)
    return inner



# TODO: This registry is a bit awful. I probably need another
# abstraction from models -> types. The type would store the graphql
# type, create input, update input, etc. This way I can map from model
# to type, and then access the graphql type I need.

graphql_field_registry = {
}

graphql_type_registry = {
}

graphql_create_input_registry = {
}

graphql_update_input_registry = {
}

graphql_reverse_input_registry = {
}


class FieldMetaclass(type):
    def __init__(cls, name, bases, dct):
        if bases and name != 'Field':
            for src in getattr(cls, 'sources', ()):
                graphql_field_registry[src] = cls
        super().__init__(name, bases, dct)


def add_graphql_field_mapping(source, field_class):
    # TODO: Uniqueness
    graphql_field_registry[source] = field_class


def add_graphql_type(model, type):
    graphql_type_registry[model] = type


def add_graphql_create_input(model, type):
    graphql_create_input_registry[model] = type


def add_graphql_update_input(model, type):
    graphql_update_input_registry[model] = type



from functools import wraps


def decorator_with_args(func):
    @wraps(func)
    def inner(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return func(args[0])
        else:
            return lambda f: func(f, *args, **kwargs)
    return inner



from polecat.db.session import Session


class Event:
    def __init__(self, event):
        self.event = event
        self.session = Session()

    def is_http(self):
        return False

    def is_admin(self):
        return self.event.get('event') == 'admin'


class HttpEvent(Event):
    def __init__(self, event, request=None):
        super().__init__(event)
        self.request = request or event

    def is_http(self):
        return True

    def get_authorization_header(self):
        return self.request.headers.get('authorization', None)



from polecat.utils import to_tuple
from polecat.utils.singledict import SingleDict

from .model import Mutation, Type


class Blueprint:
    def __init__(self):
        self.models = SingleDict()
        self.types = SingleDict()
        self.roles = SingleDict()
        self.access = []
        self.queries = SingleDict()
        self.mutations = SingleDict()
        self.hooks = []

    def add_model(self, model):
        self.models[model.Meta.name] = model
        self.add_model_hooks(model)

    def add_model_hooks(self, model):
        self.add_model_field_hooks(model)
        hooks = to_tuple(getattr(model.Meta, 'hooks', ()))
        context = HookContext(self, model_class=model)
        for hook in hooks:
            self.add_hook(hook, context)

    def add_model_field_hooks(self, model):
        for field_name, field in model.Meta.fields.items():
            hooks = to_tuple(getattr(field, 'hooks', ()))
            context = HookContext(
                self, model_class=model, field=field, field_name=field_name
            )
            for hook in hooks:
                self.add_hook(hook, context)

    def add_type(self, type):
        self.types[type.Meta.name] = type

    def add_role(self, role):
        # TODO: Should really be using name...
        self.roles[role.Meta.role] = role

    def add_access(self, access):
        self.access.append(access)

    def add_query(self, query):
        self.queries[query.Meta.name] = query

    def add_mutation(self, mutation):
        self.mutations[mutation.name] = mutation

    def create_type(self, name, fields):
        type(name, (Type,), {**fields})

    def create_mutation(self, name, resolver, return_type, input_type=None):
        cls = type(name, (Mutation,), {
            'returns': return_type,
            'input': input_type
        })
        cls.resolve = classmethod(resolver)

    def add_hook(self, callback, context=None):
        self.hooks.append((callback, context))

    def run_hooks(self):
        default_context = HookContext(self)
        for hook, context in self.hooks:
            if context is None:
                context = default_context
            hook(context)

    def get_mutation(self, name):
        return self.mutations[name]

    def iter_models(self):
        return self.models.values()

    def iter_types(self):
        return self.types.values()

    def iter_roles(self):
        return self.roles.values()

    def iter_access(self):
        return self.access

    def iter_queries(self):
        return self.queries.values()

    def iter_mutations(self):
        return self.mutations.values()


class HookContext:
    def __init__(self, blueprint, model_class=None, field=None,
                 field_name=None):
        self.blueprint = blueprint
        self.model_class = model_class
        self.field = field
        self.field_name = field_name



from ...utils.repr import to_repr
from .entity import AsIs, ConstructionArguments, Entity
from .utils import column_to_identifier, table_to_identifier

__all__ = ('Column', 'MutableColumn', 'IntColumn', 'TextColumn',
           'BoolColumn', 'FloatColumn', 'RelatedColumn', 'ReverseColumn',
           'PasswordColumn', 'DateColumn', 'TimestampColumn', 'UUIDColumn', 'SerialColumn',
           'PointColumn', 'JSONColumn')


class Column(Entity):
    def __init__(self, name, unique=False):
        self.name = name
        self.unique = unique

    def __repr__(self):
        return to_repr(
            self,
            name=self.name
        )

    @property
    def signature(self):
        return (Column, self.name)

    def has_changed(self, other):
        return (
            self.__class__ != other.__class__ or
            self.name != other.name
        )

    def bind(self, table):
        self.table = table

    def to_db_value(self, value):
        return value

    def from_db_value(self, value):
        return value

    def get_construction_arguments(self):
        return ConstructionArguments(
            self.name,
            unique=self.unique
        )


class MutableColumn(Column):
    def __init__(self, name, null=False, primary_key=False, default=None,
                 unique=False, dimensions=None, **kwargs):
        super().__init__(name, unique=True if primary_key else unique, **kwargs)
        self.null = False if primary_key else null
        self.primary_key = primary_key
        self.default = default
        self.dimensions = dimensions

    def __repr__(self):
        return to_repr(
            self,
            name=self.name
        )

    def has_changed(self, other):
        return (
            super().has_changed(other) or
            self.null != other.null or
            self.primary_key != other.primary_key or
            self.default != other.default
        )

    def get_construction_arguments(self):
        cargs = super().get_construction_arguments()
        return cargs.merge(
            unique=self.unique,
            null=self.null,
            primary_key=self.primary_key,
            default=self.default,
            dimensions=self.dimensions
        )


class IntColumn(MutableColumn):
    pass


class SerialColumn(MutableColumn):
    pass


class FloatColumn(MutableColumn):
    pass


class PointColumn(MutableColumn):
    pass


class BoolColumn(MutableColumn):
    pass


class TextColumn(MutableColumn):
    def __init__(self, *args, max_length=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_length = max_length

    def has_changed(self, other):
        return (
            super().has_changed(other) or
            self.max_length != other.max_length
        )

    def get_construction_arguments(self):
        cargs = super().get_construction_arguments()
        return cargs.merge(
            max_length=self.max_length
        )


class PasswordColumn(TextColumn):
    pass


class DateColumn(MutableColumn):
    pass


class TimestampColumn(MutableColumn):
    pass


class UUIDColumn(MutableColumn):
    pass


class JSONColumn(MutableColumn):
    pass


class RelatedColumn(IntColumn):
    def __init__(self, name, related_table, *args, related_column=None,
                 on_delete=None, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.related_table = related_table
        self.related_column = related_column
        self.on_delete = on_delete

    def __repr__(self):
        return to_repr(
            self,
            name=self.name,
            related_table=self.related_table.name
        )

    def get_dependent_entities(self):
        return (self.related_table,)

    def has_changed(self, other):
        return (
            super().has_changed(other) or
            self.related_table.signature != other.related_table.signature or
            self.related_column != other.related_column
        )

    def bind(self, table):
        super().bind(table)
        if self.should_bind_related_table():
            self.bind_related_table()
        if self.should_bind_related_column():
            self.bind_related_column(table)

    def should_bind_related_table(self):
        return isinstance(self.related_table, str)

    def bind_related_table(self):
        self.related_table = self.table.schema.get_table_by_name(
            self.related_table
        )

    def should_bind_related_column(self):
        return self.related_column and isinstance(self.related_column, str)

    def bind_related_column(self, table):
        self.related_column = ReverseColumn(
            self.related_column,
            table,
            self
        )
        self.related_table.add_column(self.related_column)

    def get_construction_arguments(self):
        cargs = super().get_construction_arguments()
        kwargs = {
            'related_table': table_to_identifier(self.related_table),
            'related_column': column_to_identifier(self.related_column),
        }
        if self.on_delete:
            kwargs['on_delete'] = AsIs(f'schema.{self.on_delete}')
        return cargs.merge(**kwargs)


class ReverseColumn(Column):
    def __init__(self, name, related_table, related_column):
        self.name = name
        self.related_table = related_table
        self.related_column = related_column

    def __repr__(self):
        return to_repr(
            self,
            name=self.name,
            related_table=self.related_table
        )

    def has_changed(self, other):
        return (
            super().has_changed(other) or
            self.related_table.signature != other.related_table.signature or
            self.related_column.signature != other.related_column.signature
        )



import re


class DevelopMiddleware:
    host_prog = re.compile(r'[a-zA-Z0-9._-]+\.amazonaws\.com')

    def run(self, event, context=None):
        match = self.host_prog.match(event.request.host or '')
        if match:
            if event.request.path:
                ii = 0
                if event.request.path[0] == '/':
                    ii += 1
                ii = event.request.path.find('/', ii)
                event.request.path = event.request.path[ii:]



import click

from .main import main

__all__ = ('server',)


@main.command()
@click.option('--production', is_flag=True, default=False,
              help='Launch in production mode.')
def server(production):
    from ..deploy.server.server import Server
    server = Server(production=production)
    server.run()



import re

import ujson
from stdb.core.engine import default_engine as engine
from stdb.core.function import default_function_registry as function_registry
from stdb.utils import to_bool


class Filter:
    FILTER_PROG = re.compile(r'^([a-zA-Z]+(?:__[a-zA-Z]+)*)$')
    FILTER_TYPES = None

    def __init__(self, args):
        self.root = self.parse_args(args)

    def get_sql(self, version, endpoint):
        self.version = version
        self.endpoint = endpoint
        if self.root:
            return self.root.get_sql(self)
        else:
            return None

    def parse_args(self, args):
        root = None
        if args:
            for k, v in args.items():
                m = self.FILTER_PROG.match(k)
                if not m:
                    continue
                target = m.group(1)
                lookup, flt_cls = self.parse_target(target)
                flt = flt_cls(self, lookup, v)
                if root is None:
                    root = flt
                else:
                    root = And(root, flt)
        return root

    def parse_target(self, target):
        i = target.rfind('__')
        if i != -1:
            try:
                return target[:i], self.FILTER_TYPES[target[i + 2:]]
            except KeyError:
                pass
        return target, Equal

    def merge(self, other):
        # TODO: We should really do a check for duplicate filters.
        if self.root:
            if other.root:
                self.root = And(self.root, other.root)
        elif other.root:
            self.root = other.root


class FilterType:
    def __init__(self, filter, lookup, value):
        self.parse_lookup(lookup)
        self.parse_value(filter, value)

    def get_sql(self, filter):
        sql = self.eval(filter)
        sql = self.eval_joins(filter, sql)
        return sql

    def eval(self, filter):
        pass
        # val = self.value
        # if isinstance(self.value, str):
        #     val = val.format(**filter.context)
        # values.append(val)

    def eval_joins(self, filter, condition):
        if not self.joins:
            return condition
        sql = '{next}'
        ep = filter.endpoint
        for j in self.joins:
            # TODO: Handle m2m, reverse fk, reverse m2m.
            prev_tbl = ep['table']
            prev_col = ep['fields'][j]['column']
            try:
                ep = filter.version[ep['fields'][j]['field']['other']]
            except KeyError:
                raise ValueError(f'invalid include during lookup: {j}')
            tbl = ep['table']
            next = 'EXISTS (SELECT 1 FROM %s WHERE %s = %s AND {next})' % (
                tbl,
                f'{prev_tbl}.{prev_col}',
                f'{tbl}.id'
            )
            sql = sql.format(next=next)
        sql = sql.format(next=condition)
        return sql

    def parse_lookup(self, lookup):
        lookup_parts = lookup.split('__')
        if len(lookup_parts) < 1:
            raise ValueError(f'invalid filter: {lookup}')
        # if lookup_parts[-1] in Filter.FILTER_TYPES:
        #     self.type = lookup_parts.pop()
        # else:
        #     self.type = 'eq'
        self.joins = lookup_parts[:-1]
        self.field = lookup_parts[-1]

    def parse_value(self, filter, value):
        if isinstance(value, str) and len(value) and value[0] == '#':
            func = function_registry.parse(value[1:])
            value = func()
        self.value = engine.quote_literal(value)

    def get_table_column(self, filter):
        ep = filter.endpoint
        for j in self.joins:
            try:
                ep = filter.version[ep['fields'][j]['field']['other']]
            except KeyError:
                raise ValueError(f'invalid include during lookup: {j}')
        if self.field == 'id':
            return ep['table'], 'id'
        else:
            try:
                return ep['table'], ep['fields'][self.field]['column']
            except KeyError:
                raise ValueError(f'invalid field during lookup: {self.field}')


class Equal(FilterType):
    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        op = '=' if self.value is not None else 'IS'
        return f'{tbl}.{col} {op} %s' % self.value


class NotEqual(FilterType):
    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        op = '!=' if self.value is not None else 'IS NOT'
        return f'{tbl}.{col} {op} %s' % self.value


class Contains(FilterType):
    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        return f'{tbl}.{col} LIKE %s' % self.value

    def parse_value(self, filter, value):
        value = '%{}%'.format(value)
        self.value = engine.quote_literal(value.replace('%', r'%%'))


class Less(FilterType):
    def eval(self, filter):
        super().eval(filter)
        return f'{self.field} < %s' % self.value


class Greater(FilterType):
    def eval(self, filter):
        super().eval(filter)
        return f'{self.field} > %s' % self.value


class LessEqual(FilterType):
    def eval(self, filter):
        super().eval(filter)
        return f'{self.field} <= %s' % self.value


class GreaterEqual(FilterType):
    def eval(self, filter):
        super().eval(filter)
        return f'{self.field} >= %s' % self.value


class In(FilterType):
    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        return f'{tbl}.{col} = ANY (%s)' % self.value

    def parse_value(self, filter, value):
        try:
            self.value = engine.quote_array(ujson.loads(value))
        except Exception:
            raise ValueError(f'Unable to parse "in" filter value: {value}')


class NotIn(In):
    def eval(self, filter):
        FilterType.eval(self, filter)
        return f'{self.field} NOT IN %s' % self.value


class IsNull(FilterType):
    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        op = 'IS' if self.value else 'IS NOT'
        return f'{tbl}.{col} {op} NULL'

    def parse_value(self, filter, value):
        self.value = to_bool(value)


class NotNull(FilterType):
    def eval(self, filter):
        super().eval(filter)
        return f'{self.field} NOTNULL'


class Overlap(FilterType):
    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        return f'{tbl}.{col} && {self.value}'


class Operator:
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def get_sql(self, filter):
        raise NotImplementedError

    def eval_sides(self, filter):
        left = self.left.get_sql(filter)
        right = self.right.get_sql(filter)
        return left, right


class And(Operator):
    def get_sql(self, filter):
        left, right = self.eval_sides(filter)
        if isinstance(self.left, Or):
            left = f'({left})'
        if isinstance(self.right, Or):
            right = f'({right})'
        return f'{left} AND {right}'


class Or(Operator):
    def get_sql(self, filter):
        left, right = self.eval_sides(filter)
        return f'{left} OR {right}'


Filter.FILTER_TYPES = {
    'eq': Equal,
    'ne': NotEqual,
    'lt': Less,
    'gt': Greater,
    'le': LessEqual,
    'ge': GreaterEqual,
    'in': In,
    'ct': Contains,
    'ni': NotIn,
    'nu': IsNull,
    'nn': NotNull,
    'ov': Overlap,
    # 'bt': Between
}



# from .filter import Filter


class Query:
    def __init__(self, model):
        self.model = model
        self.page_size = 20

    def iter_rows(self):
        raise NotImplementedError

    def get_where_sql(self, model):
        flt = getattr(self, 'filter', None)
        # if 'default_filter' in endpoint:
        #     def_flt = Filter(endpoint['default_filter'])
        #     if flt:
        #         flt.merge(def_flt)
        #     else:
        #         flt = def_flt
        if flt:
            flt = self.filter.get_sql(self.version, self.endpoint)
            if flt:
                return f'WHERE {flt}'
        return ''



from psycopg2.sql import SQL, Identifier

from .expression import Expression


class LocalVariable(Expression):
    def __init__(self, key, value):
        self.key = key
        self.value = value

    def to_sql(self):
        return (
            SQL('SET LOCAL {} TO %s').format(
                Identifier(self.key)
            ),
            (self.value,)
        )


class LocalRole(Expression):
    def __init__(self, role):
        self.role = role

    def to_sql(self):
        return (
            SQL('SET LOCAL ROLE {}').format(
                Identifier(self.role)
            ),
            ()
        )



from polecat.db.query.query import Values as QueryValues
from polecat.db.sql.expression import (LocalRole, LocalVariable, Multi, Select,
                                       Values)

from ..schema import create_table


def test_multipart_expression(immutabledb):
    table = create_table()
    expr = Multi(
        LocalRole('role'),
        LocalVariable('a', 1),
        Select(table, ('col1',))
    )
    sql = immutabledb.mogrify(*expr.to_sql())
    assert sql == (
        b'SET LOCAL ROLE "role"; SET LOCAL "a" TO 1; SELECT "a_table"."col1"'
        b' AS "col1" FROM "a_table"'
    )


def test_values_with_dict(immutabledb):
    expr = Values({
        'a': 1,
        'b': 2
    })
    sql = immutabledb.mogrify(*expr.to_sql())
    assert sql == b'VALUES (1, 2)'


def test_values_with_query_values(immutabledb):
    expr = Values(
        QueryValues({
            'a': 1,
            'b': 2
        })
    )
    sql = immutabledb.mogrify(*expr.to_sql())
    assert sql == b'VALUES (1, 2)'


def test_values_with_bulk_query_values(immutabledb):
    expr = Values(
        QueryValues(
            (
                (1, 2),
                (3, 4)
            ),
            ('a', 'b')
        )
    )
    sql = immutabledb.mogrify(*expr.to_sql())
    assert sql == b'VALUES (1, 2), (3, 4)'



class Executor:
    def __init__(self, expression, cursor):
        self.expression = expression
        self.cursor = cursor

    def execute(self, cursor):
        sql, args = self.expression.to_sql()
        cursor.execute(sql, args)
        for row in cursor:
            yield row



from polecat.utils import to_list

from .exceptions import InvalidFieldError
from .omit import NONE
from .registry import (AccessMetaclass, ModelMetaclass, MutationMetaclass,
                       QueryMetaclass, RoleMetaclass, TypeMetaclass)


# TODO: Types and models are soooo similar, perhaps use a better
# architecture than total separation?
class Type(metaclass=TypeMetaclass):
    _registry = 'type_registry'
    _registry_base = 'Type'

    def __init__(self, **kwargs):
        for field_name, value in kwargs.items():
            setattr(self, field_name, value)

    def __repr__(self):
        return f'<{self.Meta.name}>'


class Model(metaclass=ModelMetaclass):
    def __init__(self, **kwargs):
        for field_name, value in kwargs.items():
            try:
                field = self.Meta.fields[field_name]
            except KeyError:
                raise InvalidFieldError(self.__class__, field_name)
            setattr(self, field_name, field.from_incoming(self, value))

    def __repr__(self):
        return f'<{self.Meta.name}>'


class Role(metaclass=RoleMetaclass):
    @classmethod
    def db_name(self):
        # TODO: This is the worst.
        from polecat.project.project import get_active_project
        # TODO: Plus, this whole thing is duplicated from "db". I need
        # to be able to map from these declarative roles to the DB
        # ones. Probably just create the DB one during preparation?
        name = ''
        project = get_active_project()
        if project:
            name += project.name
        if self.Meta.app:
            name += self.Meta.app.name
        if name:
            name += '_'
        name += self.Meta.role
        return name


class Access(metaclass=AccessMetaclass):
    pass


class Query(metaclass=QueryMetaclass):
    pass


class Mutation(metaclass=MutationMetaclass):
    def resolve(self, context):
        raise NotImplementedError



import ujson
from datetime import date, datetime
from uuid import UUID

from graphql.error import INVALID
from graphql.language.ast import (BooleanValueNode, FloatValueNode,
                                  IntValueNode, StringValueNode)
from graphql.pyutils import inspect, is_finite, is_integer
from graphql.type.definition import GraphQLScalarType, is_named_type

__all__ = ('GraphQLDatetime', 'GraphQLDate', 'GraphQLJSON')


def serialize_date(value):
    if isinstance(value, str):
        # TODO: Confirm value is a date?
        return value
    if isinstance(value, date):
        # TODO: Which format?
        return str(value)
    # Do not serialize builtin types as strings, but allow
    # serialization of custom types via their `__str__` method.
    if type(value).__module__ == 'builtins':
        raise TypeError(f'Date cannot represent value: {inspect(value)}')
    return str(value)


def coerce_date(value):
    if not isinstance(value, str):
        raise TypeError(
            f'Date cannot represent a non string value: {inspect(value)}'
        )
    return value


def parse_date_literal(ast, _variables=None):
    """ Parse a string value node in the AST.
    """
    if isinstance(ast, StringValueNode):
        # TODO: Must be a datetime.
        return ast.value
    return INVALID


GraphQLDate = GraphQLScalarType(
    name="Date",
    description="The `Date` scalar type represents"
    " a date.",
    serialize=serialize_date,
    parse_value=coerce_date,
    parse_literal=parse_date_literal
)


def serialize_datetime(value):
    if isinstance(value, str):
        # TODO: Confirm value is a datetime?
        return value
    if isinstance(value, datetime):
        # TODO: Which format?
        return str(value)
    # Do not serialize builtin types as strings, but allow
    # serialization of custom types via their `__str__` method.
    if type(value).__module__ == 'builtins':
        raise TypeError(f'Datetime cannot represent value: {inspect(value)}')
    return str(value)


def coerce_datetime(value):
    if not isinstance(value, str):
        raise TypeError(
            f'Datetime cannot represent a non string value: {inspect(value)}'
        )
    return value


def parse_datetime_literal(ast, _variables=None):
    """ Parse a string value node in the AST.
    """
    if isinstance(ast, StringValueNode):
        # TODO: Must be a datetime.
        return ast.value
    return INVALID


GraphQLDatetime = GraphQLScalarType(
    name="Datetime",
    description="The `Datetime` scalar type represents"
    " a date and time containing a timezone.",
    serialize=serialize_datetime,
    parse_value=coerce_datetime,
    parse_literal=parse_datetime_literal,
)


def serialize_uuid(value):
    if isinstance(value, str):
        # TODO: Confirm value is a UUID?
        return value
    if isinstance(value, UUID):
        # TODO: Which format?
        return str(value)
    # Do not serialize builtin types as strings, but allow
    # serialization of custom types via their `__str__` method.
    if type(value).__module__ == 'builtins':
        raise TypeError(f'UUID cannot represent value: {inspect(value)}')
    return str(value)


def coerce_uuid(value):
    if not isinstance(value, str):
        raise TypeError(
            f'UUID cannot represent a non string value: {inspect(value)}'
        )
    return value


def parse_uuid_literal(ast, _variables=None):
    """ Parse a string value node in the AST.
    """
    if isinstance(ast, StringValueNode):
        # TODO: Must be a UUID.
        return ast.value
    return INVALID


GraphQLUUID = GraphQLScalarType(
    name="UUID",
    description="The `UUID` scalar type represents"
    " a UUID.",
    serialize=serialize_uuid,
    parse_value=coerce_uuid,
    parse_literal=parse_uuid_literal,
)


def serialize_json(value):
    return ujson.dumps(value)


def coerce_json(value):
    return ujson.loads(value)


def parse_json_literal(ast, _variables=None):
    """ Parse a string value node in the AST.
    """
    if isinstance(ast, StringValueNode):
        # TODO: Must be a UUID.
        return ast.value
    return INVALID


GraphQLJSON = GraphQLScalarType(
    name="JSON",
    description="The `JSON` scalar type represents"
    " a JSON.",
    serialize=serialize_json,
    parse_value=coerce_json,
    parse_literal=parse_json_literal,
)


# TODO: Prefer tuple, but we add it to a list in `./schema.py`.
scalars = [GraphQLDatetime, GraphQLUUID, GraphQLJSON]



from botocore.exceptions import ClientError, WaiterError

import ujson
from termcolor import colored

from ...utils import get_path, set_path
from ...utils.feedback import feedback
from .constants import PROJECT_PREFIX
from .deployment import assert_deployment_exists
from .exceptions import KnownError
from .operations import get_parameter, get_parameters_by_path
from .resources import (create_api_resources, create_domain_resources,
                        create_output_resources, create_zone_resources)
from .utils import aws_client

# TODO: Need to make the API name dynamic.
api_name = 'Polecat1'


@feedback
def deploy(project, bucket, deployment=None, dry_run=False, feedback=None):
    cf = aws_client('cloudformation')
    msg = f'Deploy {colored(project, "blue")}'
    if deployment:
        msg += f'/{colored(deployment, "green")}'
    with feedback(msg):
        template = create_template(project, bucket, deployment=deployment, cf=cf)
        if dry_run:
            raise Warning('dry run')
        params = {
            'StackName': project,
            'TemplateBody': template,
            'Capabilities': ['CAPABILITY_IAM']
        }
        wait_command = 'stack_update_complete'
        try:
            cf.update_stack(**params)
        except ClientError as e:
            if 'No updates are to be performed' in str(e):
                raise Warning('no changes detected')
            elif 'does not exist' not in str(e):
                raise
            wait_command = 'stack_create_complete'
            cf.create_stack(**params)
        waiter = cf.get_waiter(wait_command)
        try:
            waiter.wait(StackName=project)
        except WaiterError:
            raise KnownError('failed')
    response = cf.describe_stacks(StackName=project)
    outputs = {}
    for output in response['Stacks'][0]['Outputs']:
        project_deployment = output['OutputKey']
        deployment = project_deployment[len(project):-3].lower()
        url = output['OutputValue']
        outputs[deployment] = url
    return outputs


@feedback
def undeploy(project, deployment=None, feedback=None):
    with feedback(f'Undeploy {colored(project, "blue")}'):
        cf = aws_client('cloudformation')
        cf.delete_stack(StackName=project)
        # TODO: Wait for delete to be finished.
        # TODO: Undeploy individual deployment?


def create_template(project, bucket, deployment=None, cf=None):
    env = get_environment(project, deployment)
    resources, outputs = get_existing_template(project, cf=cf)
    data = {
        'AWSTemplateFormatVersion': '2010-09-09',
        'Description': f'Polecat {project} CloudFormation template',
        'Resources': create_resources(project, bucket, env, resources=resources),
        'Outputs': create_outputs(project, env, outputs=outputs)
    }
    return ujson.dumps(data, indent=2)


def get_existing_template(project, cf=None):
    cf = aws_client('cloudformation', client=cf)
    try:
        response = cf.get_template(StackName=project, TemplateStage='Original')
        template = response['TemplateBody']
        return template['Resources'], template['Outputs']
    except ClientError as e:
        if 'does not exist' not in str(e):
            raise
        return {}, {}


def create_resources(project, bucket, environment, resources=None):
    resources = resources or {}
    for dep, env in environment.items():
        resources.update(create_api_resources(project, dep, api_name, bucket, env))
        for domain, info in env.get('domains', {}).items():
            certificate_arn = info['certificate']  # TODO: Confirm it exists?
            resources.update(
                create_domain_resources(project, dep, api_name, domain, certificate_arn)
            )
            zone = info.get('zone', None)
            if zone:
                resources.update(
                    create_zone_resources(project, dep, domain, zone)
                )
    return resources


def create_outputs(project, environment, outputs=None):
    outputs = outputs or {}
    for dep, env in environment.items():
        outputs.update(create_output_resources(project, dep, api_name))
    return outputs


def get_environment(project, deployment=None):
    deployment = deployment or ''
    if deployment:
        assert_deployment_exists(project, deployment)
    prefix = PROJECT_PREFIX.format(project)
    project_code_version = get_parameter(f'{prefix}code/version')  # TODO: Constant.
    if project_code_version is None:
        raise KnownError('No code version available')
    project_bundle_version = get_parameter(f'{prefix}bundle/version', default='')  # TODO: Constant.
    path = f'/polecat/projects/{project}/deployments/{deployment}' + ('/' if deployment else '')
    environment = get_parameters_by_path(path)
    if deployment:
        environment = {
            deployment: environment
        }
    if project_code_version:
        for dep, env in environment.items():
            if get_path(env, 'code.version') is None:
                set_path(env, 'code.version', project_code_version)
            if get_path(env, 'bundle.version') is None:
                set_path(env, 'bundle.version', project_bundle_version)
    return environment



from psycopg2.sql import SQL

from .expression import Expression


class Union(Expression):
    def __init__(self, expressions=None):
        self.expressions = expressions or []

    def add_expression(self, expression):
        self.expressions.append(expression)

    def to_sql(self):
        all_sql = []
        all_args = ()
        for expr in self.expressions:
            sql, args = expr.to_sql()
            all_sql.append(sql)
            all_args += args
        return SQL(' UNION ALL ').join(all_sql), all_args

    def push_selection(self, selection):
        for expr in self.expressions:
            expr.push_selection(selection)



import os
import shlex
import subprocess
from tempfile import TemporaryDirectory
from zipfile import ZipFile

import click
import ujson

from .main import main

__all__ = ('test_production', 'test_with_pytest')

api_gateway_request = {
    'path': '/graphql',
    'httpMethod': 'GET',
    'headers': {},
    'body': '{}'
}


@main.group()
@click.pass_context
def test(ctx):
    pass


@test.command('production')
@click.argument('source')
@click.pass_context
def test_production(ctx, source):
    with TemporaryDirectory() as root:
        os.chmod(root, 0o755)
        with ZipFile(source, 'r') as zf:
            zf.extractall(root)
        env = {
            'DATABASE_URL': os.environ['DATABASE_URL'],
            'POLECAT_PROJECT_MODULE': 'basic.project.Project'  # TODO
        }
        env_str = ' '.join(f'-e {k}={v}' for k, v in env.items())
        request = api_gateway_request
        cmd = (
            f'docker run --rm -v {root}:/var/task {env_str}'
            f' lambci/lambda:python3.7 main.handler \'{ujson.dumps(request)}\''  # TODO: Escape
        )
        subprocess.call(shlex.split(cmd))


@test.command('pytest')
@click.pass_context
def test_with_pytest(ctx):
    cmd = 'pytest'
    subprocess.call(shlex.split(cmd))



from contextlib import contextmanager

import boto3

from .exceptions import EntityDoesNotExist, EntityExists


def aws_client(service, client=None, region=None):
    return client or boto3.client(service, region_name=region)


def aws_resource(service, resource=None):
    return resource or boto3.resource(service)


@contextmanager
def aws_exists():
    try:
        yield
    except Exception as e:
        try:
            code = e.response['Error']['Code']
        except (KeyError, AttributeError):
            code = None
        if code in ('EntityAlreadyExists', 'BucketAlreadyOwnedByYou',
                    'BucketPolicyAlreadyExists', 'ParameterAlreadyExists'):
            raise EntityExists
        raise


@contextmanager
def aws_does_not_exist(swallow=False):
    try:
        yield
    except Exception as e:
        try:
            code = e.response['Error']['Code']
        except (KeyError, AttributeError):
            code = None
        if code in ('DBInstanceNotFound',):
            if not swallow:
                raise EntityDoesNotExist
        else:
            raise


def default_path(path):
    return path or '/polecat/'


def assert_project_and_deployment_exists(project, deployment, ssm=None):
    from .project import assert_project_exists
    from .deployment import assert_deployment_exists
    assert_project_exists(project, ssm=ssm)
    assert_deployment_exists(deployment, ssm=ssm)



import click

from .main import main

__all__ = ('admin',)


@main.group()
@click.option('--deployment', '-d')
@click.pass_context
def admin(ctx, deployment):
    if deployment:
        ctx.obj['deployment'] = deployment



from psycopg2.sql import SQL, Identifier


class Policy:
    def __init__(self, name, expression):
        self.name = name.replace('.', '_')
        self.expression = expression

    def bind(self, table):
        self.table = table
        self.full_name = '{}_{}_pol'.format(
            self.table.name,
            self.name
        )

    @property
    def sql(self):
        sql = SQL('')
        if not self.table.is_rls_enabled():
            sql += SQL('ALTER TABLE {} ENABLE ROW LEVEL SECURITY; ').format(
                Identifier(self.table.name)
            )
        sql += SQL('CREATE POLICY {} ON {} USING ({})').format(
            Identifier(self.full_name),
            Identifier(self.table.name),
            SQL(self.expression)
        )
        return sql, ()

    def serialize(self):
        return 'policy.Policy(name={}, expression={})'.format(
            repr(self.name),
            repr(self.expression)
        )



import re
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path

from polecat.db.schema import Schema

from ...utils import indent
from ..connection import transaction
from ..decorators import dbcursor
from .utils import project_migrations_path

migration_template = '''from polecat.db.migration import migration, operation
from polecat.db import schema
from polecat.db.schema import column, index, policy


class Migration(migration.Migration):
    dependencies = [{}]
    operations = [{}]
'''


class Migration:
    filename_prog = re.compile(r'(\d{4})_[a-zA-Z_0-9]+\.py')

    @classmethod
    def load_migration_class(cls, path, app=None):
        module_path = path.name[:-3]
        if app:
            module_path = f'{app}.{module_path}'
        spec = spec_from_file_location(module_path, str(path))
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
        return module.Migration

    def __init__(self, operations=None, app=None, dependencies=None, name=None, filename=None):
        self.operations = operations or getattr(self, 'operations', [])
        self.app = app
        self.name = name
        self.dependencies = dependencies or getattr(self, 'dependencies', [])
        self._filename = filename

    @dbcursor
    def forward(self, schema=None, migrations=None, schema_only=False, cursor=None):
        # TODO: This "_applied" thing means we can only run this
        # once. Probably need to improve that.
        if not getattr(self, '_applied', False):
            self._applied = True
            if not schema:
                schema = Schema()
            if not schema_only:
                if self.app:
                    args = [self.app.name]
                else:
                    args = []
                args.append(self.name)
                sql = (
                    'SELECT EXISTS('
                    '  SELECT 1 FROM polecat_migrations'
                    '    WHERE app {}'
                    '      AND name = %s'
                    ');'
                ).format(
                    '= %s' if self.app else 'IS NULL'
                )
                cursor.execute(sql, args)
                is_applied = cursor.fetchone()[0]
            else:
                is_applied = False
            migrations = migrations or {}
            for dep in self.dependencies:
                if isinstance(dep, str):
                    dep = migrations[dep]
                dep.forward(schema, migrations, cursor=cursor)
            # TODO: This transaction is slightly inefficient.
            with transaction(cursor):
                for op in self.operations:
                    op.forward_schema(schema)
                    if not is_applied and not schema_only:
                        op.forward(schema, cursor=cursor)
                if not is_applied and not schema_only:
                    sql = (
                        'INSERT INTO polecat_migrations (app, name, applied)'
                        '  VALUES(%s, %s, now());'
                    )
                    cursor.execute(sql, (self.app.name if self.app else None, self.name))

    @dbcursor
    def forward_post_ops(self, cursor=None):
        with transaction(cursor):
            for op in self.operations:
                for sql, args in op.post_ops:
                    # TODO: This isn't a nice check.
                    if sql.seq:
                        cursor.execute(sql, args)

    @property
    def filename(self):
        if not self._filename:
            existing_migrations = self.find_existing_migrations()
            if existing_migrations:
                next_number = existing_migrations[-1][0] + 1
            else:
                next_number = 1
            self._filename = f'{next_number:04}_migration.py'
        return self._filename

    @property
    def migrations_path(self):
        try:
            return self.app.path / 'migrations'
        except AttributeError:
            return project_migrations_path()

    @property
    def dependency_string(self):
        return f'{self.app.name}.{self.filename[:-3]}'

    def set_app(self, app):
        self.app = app
        for op in self.operations:
            op.set_app(app)

    def get_file_path(self, root=None):
        if not root:
            root = self.migrations_path
        else:
            root = Path(root)
            if self.app:
                root /= self.app.name
            root /= 'migrations'
        return root / self.filename

    def save(self, output_path=None):
        if not getattr(self, '_saved', False):
            self._saved = True
            file_path = self.get_file_path(output_path)
            file_path.parent.mkdir(parents=True, exist_ok=True)
            with open(file_path, 'w') as f:
                f.write(self.serialize())
            file_path = Path(file_path).parent / '__init__.py'
            if not file_path.exists():
                file_path.touch()

    def serialize(self):
        dep_strs = []
        for dep in self.dependencies:
            # dep.save()
            dep_strs.append(f"'{dep.dependency_string}'")
        dep_strs = ',\n'.join(dep_strs)
        if dep_strs:
            dep_strs = indent(f'\n{dep_strs}', 8)
            dep_strs += indent('\n')
        op_strs = []
        for op in self.operations:
            op_strs.append(op.serialize())
        op_strs = ',\n'.join(op_strs)
        if op_strs:
            op_strs = indent(f'\n{op_strs}', 8)
            op_strs += indent('\n')
        return migration_template.format(dep_strs, op_strs)

    def find_existing_migrations(self):
        path = self.migrations_path
        migrations = []
        try:
            for filename in path.iterdir():
                match = self.filename_prog.match(filename.name)
                if match:
                    migrations.append((int(match.group(1)), filename.name))
        except FileNotFoundError:
            pass
        return sorted(migrations, key=lambda x: x[0])



class InvalidFieldError(Exception):
    def __init__(self, model, field_name):
        super().__init__(f'{model.Meta.name} has no field "{field_name}"')


class InvalidModelDataError(Exception):
    def __init__(self, model, data):
        super().__init__(f'data type {type(data)} invalid for {model.Meta.name}')



import click

from .feedback import HaloFeedback
from .main import main

__all__ = ('publish', 'unpublish')


@main.command()
@click.argument('deployment')
@click.argument('domain')
@click.option('--certificate')
@click.option('--zone')
@click.pass_context
def publish(ctx, deployment, domain, certificate, zone):
    from ..deploy.aws.publish import publish as aws_publish
    project = ctx.obj['project']
    aws_publish(project, deployment, domain, certificate, zone, feedback=HaloFeedback())


@main.command()
@click.argument('deployment')
@click.argument('domain')
@click.pass_context
def unpublish(ctx, deployment, domain):
    from ..deploy.aws.publish import unpublish as aws_unpublish
    project = ctx.obj['project']
    aws_unpublish(project, deployment, domain, feedback=HaloFeedback())



from ..schema import Role, Table
from .migration import Migration
from .operation import (AlterRole, AlterTable, CreateRole, CreateTable,
                        DeleteRole, DeleteTable, GrantAccessToRole,
                        GrantAccessToTable, RevokeAccess)


class Differ:
    def diff(self, from_schema, to_schema):
        self.operations = []
        self.diff_roles(from_schema, to_schema)
        self.diff_tables(from_schema, to_schema)
        self.diff_access(from_schema, to_schema)
        self.find_dependencies()
        by_app = {}
        for op in self.operations:
            by_app.setdefault(op.app, []).append(op)
        migrations = {
            app: Migration(ops, app=app)
            for app, ops in by_app.items()
        }
        for app, mgr in migrations.items():
            dep_mgrs = []
            seen_apps = set()
            for op in mgr.operations:
                for dep in op._dependencies:
                    if dep.app in seen_apps:
                        continue
                    seen_apps.add(dep.app)
                    if dep.app != app:
                        dep_mgrs.append(migrations[dep.app])
            mgr.dependencies = dep_mgrs
        return list(migrations.values())

    def diff_roles(self, from_schema, to_schema):
        created, deleted, altered = self.diff_entities(
            from_schema.roles,
            to_schema.roles
        )
        self.operations.extend([
            *[
                DeleteRole(role)
                for role in deleted
            ],
            *[
                CreateRole(role)
                for role in created
            ],
            *[
                AlterRole(to_role, from_role)
                for to_role, from_role in altered
            ]
        ])

    def diff_tables(self, from_schema, to_schema):
        created, deleted, altered = self.diff_entities(
            from_schema.tables,
            to_schema.tables
        )
        self.operations.extend([
            *[
                DeleteTable(table)
                for table in deleted
            ],
            *[
                CreateTable(table)
                for table in created
            ],
            *[
                AlterTable(to_table, from_table)
                for to_table, from_table in altered
            ]
        ])

    def diff_access(self, from_schema, to_schema):
        from_access = from_schema.merge_access()
        to_access = to_schema.merge_access()
        created, deleted, altered = self.diff_entities(from_access, to_access)
        grants = []
        for access in created:
            entity = access.entity
            if isinstance(entity, Table):
                grants.append(GrantAccessToTable(access))
            elif isinstance(entity, Role):
                grants.append(GrantAccessToRole(access))
            else:
                raise TypeError(f'Unknown schema entity: {access}')
        # TODO: Alter.
        self.operations.extend([
            *grants,
            *[
                RevokeAccess(access)
                for access in deleted
            ]
        ])

    def find_dependencies(self):
        op_map = {}
        for op in self.operations:
            op_map.setdefault(op.signature, []).append(op)
        for op in self.operations:
            op._dependencies = []
            for dep in op.dependencies:
                # TODO: Oh no... yuck.
                op._dependencies.extend(op_map.get(dep.signature, []))

    def diff_entities(self, from_entities, to_entities):
        # TODO: Could use set operations...?
        from_entities = {
            e.signature: e
            for e in from_entities
        }
        to_entities = {
            e.signature: e
            for e in to_entities
        }
        created = [
            e for e in to_entities.values()
            if e.signature not in from_entities
        ]
        deleted = [
            e for e in from_entities.values()
            if e.signature not in to_entities
        ]
        altered = [
            (e, from_entities[e.signature])
            for e in to_entities.values()
            if e.signature in from_entities and
            from_entities[e.signature].has_changed(e)
        ]
        # TODO: How to check for renames?
        return (created, deleted, altered)

    def merge_dependencies(self, operations):
        deps = []
        for op in operations:
            for dep in op.dependencies:
                if dep in deps:
                    continue
                deps.append(dep)
        return deps



import os

from .meta import MetaMetaclass
from .optiondict import OptionDict


class ConfigMetaclass(MetaMetaclass):
    def __new__(metaclass, name, bases, attrs):
        if name == 'Config':
            return type.__new__(metaclass, name, bases, attrs)
        return metaclass.build_dict_type(metaclass, name, attrs, ConfigDict)


class ConfigDict(OptionDict):
    def __init__(self, *args, prefix=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_prefix(prefix)

    def set_prefix(self, prefix):
        self.prefix = f'{prefix.upper()}' if prefix else ''
        if self.prefix and self.prefix[-1] != '_':
            self.prefix += '_'

    def get_default(self, option):
        key = f'{self.prefix}{option.key.upper()}'
        value = os.environ.get(key)
        if value is None:
            value = super().get_default(option)
        # TODO: Is this a problem?
        option.default = option.coerce(value)
        return option.default


class Config(metaclass=ConfigMetaclass):
    class Any:
        pass



from graphql.utilities.schema_printer import print_schema as schema_repr


def print_schema(schema):
    print(schema_repr(schema))


def get_model_class_from_info(info):
    return info.return_type._model



import os

from polecat.db.connection import ConnectionManager


def test_uses_environment():
    # TODO: When pytest.org comes back online, find a nicer way.
    old_val = os.environ.get('DATABASE_URL')
    try:
        os.environ['DATABASE_URL'] = 'test'
        manager = ConnectionManager()
        assert manager.stack == ['test']
    finally:
        if old_val is not None:
            os.environ['DATABASE_URL'] = old_val
        else:
            del os.environ['DATABASE_URL']


def test_push_and_pop_url():
    manager = ConnectionManager()
    with manager.push_url(os.environ['DATABASE_URL']):
        with manager.connection():
            assert list(manager.connections.keys()) != []
        assert manager.stack == 2*[os.environ['DATABASE_URL']]
    assert manager.stack == [os.environ['DATABASE_URL']]
    assert list(manager.connections.keys()) == []


def test_connection_reuse():
    url = os.environ['DATABASE_URL']
    manager = ConnectionManager()
    assert url not in manager.connections
    with manager.connection() as conn:
        assert conn.status == 1
    assert url in manager.connections
    assert manager.connections[url].status == 1
    manager.close_all_connections()
    assert url not in manager.connections



from stdb.core.engine import default_engine as engine

from .query import Query
from .coerce import coerce


class BulkUpdate(Query):
    def __init__(self, version, endpoint, data):
        super().__init__(version, endpoint)
        self.data = data

    def execute(self):
        sql = self.get_main_sql(self.data)
        if sql:
            engine.execute(sql, len(self.data))
        sql = self.get_m2m_sql(self.data)
        if sql:
            engine.execute(sql)

    def get_main_sql(self, data):
        assert isinstance(data, list)
        table = self.endpoint['table']
        fields = self.endpoint['fields']

        # Construct the SQL.
        sql = ''
        for single_data in data:
            id = single_data['id']
            ii = 0
            sub_sql = ''
            for fn in single_data:
                if fn == 'id':
                    continue
                try:
                    fld = fields[fn]
                except KeyError:
                    raise ValueError(f'invalid field name: {fn}')
                if fld.get('related', False):
                    if 'm2m_table' in fld:
                        continue
                val = coerce(fld['field']['type'], single_data[fn])
                val = engine.quote_literal(val)
                col = fields[fn]['column']
                if ii > 0:
                    sub_sql += ','
                sub_sql += '{}={}'.format(col, val)
                ii += 1
            if sub_sql:
                sql += f'UPDATE "{table}" SET '
                sql += sub_sql
                sql += f' WHERE id={id};'

        return sql

    def get_m2m_sql(self, data):
        assert isinstance(data, list)
        fields = self.endpoint['fields']

        # Collect the set of m2m fields and the m2m table it goes with.
        m2ms = []
        for fn, fld in fields.items():
            fld = fields[fn]
            if fld.get('related', False) and 'm2m_table' in fld:
                m2ms.append((fn, fld['m2m_table']))

        # Contruct the sequence of inserts required to populate M2M
        # values.
        sql = ''
        for single_data in data:
            for fn, m2m_table in m2ms:
                val = single_data.get(fn, None)
                if not val:
                    continue
                id = single_data['id']
                if not isinstance(val, list):
                    raise ValueError('invalid many-to-many value, should be a list')
                sql += f'INSERT INTO "{m2m_table}" (lid, rid) VALUES '
                for ii, v in enumerate(val):
                    if not isinstance(v, int):
                        raise ValueError('invalid related ID, should be an integer')
                    if ii > 0:
                        sql += ','
                    sql += f'({id},{v})'
                sql += ';'

        return sql



from functools import wraps

from ..utils.container import OptionDict, passthrough

OptionDict = passthrough(OptionDict)
global_context = OptionDict()

__all__ = ('active_context',)


def active_context(*args):
    global global_context
    if len(args):
        func = args[0]
        @wraps(func)
        def inner(*args, **kwargs):
            return func(*args, context=global_context, **kwargs)
        return inner
    return global_context



from psycopg2.sql import SQL

from .expression import Expression


class ArrayAgg(Expression):
    def __init__(self, expression):
        self.expression = expression

    def to_sql(self):
        expr_sql, expr_args = self.expression.to_sql()
        return (
            SQL('array_agg({})').format(expr_sql),
            expr_args
        )



from psycopg2.sql import SQL

from .expression import Expression


class JSON(Expression):
    def __init__(self, expression):
        self.expression = expression

    def to_sql(self):
        sql, args = self.expression.to_sql()
        return SQL('SELECT row_to_json(__tl) FROM {} AS __tl').format(
            sql
        ), args

    def push_selection(self, selection=None):
        self.expression.push_selection(selection)



class Selection:
    def __init__(self, *fields, **lookups):
        self.fields = fields
        self.lookups = lookups

    def __repr__(self):
        return f'<Selector fields="{self.fields}" lookups="{self.lookups}">'

    def __iter__(self):
        for field_name in self.fields:
            yield field_name
        for field_name in self.lookups.keys():
            yield field_name

    def copy(self):
        return Selection(*self.fields, **self.lookups)

    def get(self, name, force=False):
        return self.lookups.get(name) or (Selection() if force else None)

    def merge(self, other):
        if other is not None:
            self.fields = list(set(self.fields) | set(other.fields))
            self.lookups = {**self.lookups}
            for name, sub in other.lookups.items():
                self.lookups[name] = self.lookups.get(name, Selection()).merge(sub)
        return self

    def has_lookups(self):
        return bool(self.lookups)

    def all_fields(self):
        return self.fields + tuple(self.lookups.keys())



from collections import OrderedDict


class SingleDict(OrderedDict):
    def __setitem__(self, key, value):
        if key in self:
            raise KeyError(f'Duplicate key "{key}"')
        super().__setitem__(key, value)



import os

import click

from .config import load_config


@click.group()
@click.option('--project', '-p')
@click.pass_context
def main(ctx, project):
    config, config_path = load_config()
    ctx.obj.update(config)
    if config_path:
        ctx.obj['config_path'] = os.path.relpath(config_path)
    if project:
        ctx.obj['project'] = project
    # TODO: Hmm.
    try:
        from polecat.project.project import load_project
        project = load_project()
        project.prepare()
    except ModuleNotFoundError:
        # TODO: Make this nicer.
        print('Warning: unable to find Polecat project')



import logging

from polecat.core.config import default_config

from ..connection import cursor as cursor_context  # TODO: Ugh.
from ..decorators import dbcursor
from .query import (Common, Delete, Filter, Insert, InsertIfMissing, Join,
                    Query, Select, Update, Values)
from .selection import Selection

logger = logging.getLogger(__name__)


class Q:
    def __init__(self, queryable=None, branches=None, session=None):
        self.queryable = queryable
        self.branches = branches or []
        self.session = session
        self.row_count = 0

    def __iter__(self):
        with cursor_context() as cursor:
            self.execute(cursor=cursor)
            for row in cursor:
                yield row[0]

    def __len__(self):
        # TODO: If we haven't executed, we probably should?
        return self.row_count

    def get(self):
        with cursor_context() as cursor:
            self.execute(cursor=cursor)
            if cursor.rowcount == 0:
                return None
            elif cursor.rowcount > 1:
                # TODO: Better exception.
                raise Exception('Multiple results for get query.')
            for row in cursor:
                return row[0]

    @classmethod
    def common(cls, *subqueries):
        return Q(Common(subqueries))

    @dbcursor
    def to_sql(self, cursor):
        # TODO: This is also technically outside of the scope of this
        # class.
        from ..sql.strategy import Strategy
        strategy = Strategy()
        expr = strategy.parse(self)
        return cursor.mogrify(*expr.to_sql())

    @dbcursor()
    def execute(self, cursor):
        # TODO: This is pretty bad. How to keep this purely as a
        # builder, but also inject strategy and execution knowledge
        # for convenience?
        from ..sql.strategy import Strategy
        strategy = Strategy()
        expr = strategy.parse(self)
        sql, args = expr.to_sql()
        if default_config.log_sql:
            # TODO: Get my logging sorted.
            # logger.debug(cursor.mogrify(sql, args))
            print(cursor.mogrify(sql, args))
        cursor.execute(sql, args)
        self.row_count = cursor.rowcount

    def select(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], Selection):
            selection = args[0].copy()
            if kwargs:
                selection.merge(Selection(**kwargs))
        else:
            selection = Selection(*args, **kwargs)
        return self.chain(
            Select(self.queryable, selection)
        )

    def insert(self, *subquery, **values):
        source = self.get_mutation_source()
        if len(subquery):
            if len(subquery) > 1:
                raise ValueError('Cannot pass more than one subquery')
            if len(values):
                raise ValueError('Cannot pass a subquery and keyword values')
            values = self.merge_query_branches(subquery[0])
        else:
            values = self.parse_value_dict(self.queryable, values)
        insert = Insert(source, values)
        if insert.reverse_queries:
            if self.branches is None:
                self.branches = []
            for reverse_queries in insert.reverse_queries.values():
                self.branches.extend(reverse_queries)
            # TODO: Do I need to wrap insert in a select?
        return self.chain(insert)

    def bulk_insert(self, columns, values):
        source = self.get_mutation_source()
        values = self.parse_bulk_values(columns, values)
        insert = Insert(source, values)
        if insert.reverse_queries:
            if self.branches is None:
                self.branches = []
            for reverse_queries in insert.reverse_queries.values():
                self.branches.extend(reverse_queries)
            # TODO: Do I need to wrap insert in a select?
        return self.chain(insert)

    def insert_into(self, source, **values):
        self.split_branch()
        return self.chain(
            Insert(source, values)
        )

    def insert_if_missing(self, values, defaults={}):
        source = self.get_mutation_source()
        values = self.parse_value_dict(self.queryable, values)
        defaults = self.parse_value_dict(self.queryable, defaults)
        insert = InsertIfMissing(source, values, defaults)
        # TODO: Do I need to worry about reverse queries?
        return self.chain(insert)

    def update(self, *subquery, **values):
        # TODO: Need to think about this more, but an Update is almost
        # always able to run, regardless of the mutability of the
        # source.
        source = self.queryable
        # TODO: This is duped in insert.
        if len(subquery):
            if len(subquery) > 1:
                raise ValueError('Cannot pass more than one subquery')
            if len(values):
                raise ValueError('Cannot pass a subquery and keyword values')
            values = self.merge_query_branches(subquery[0])
        else:
            values = self.parse_value_dict(self.queryable, values)
        update = Update(source, values)
        if update.reverse_queries:
            if self.branches is None:
                self.branches = []
            for reverse_queries in update.reverse_queries.values():
                self.branches.extend(reverse_queries)
            # TODO: Do I need to wrap update in a select?
        return self.chain(update)

    def update_into(self, source, **values):
        self.split_branch()
        return self.chain(Update(source, values))

    def delete(self):
        # TODO: Would be good to remove this conditional.
        if isinstance(self.queryable, Filter):
            source = self.queryable
        else:
            source = self.get_mutation_source()
        return self.chain(Delete(source))

    def delete_from(self, source):
        self.split_branch()
        return self.chain(Update(source))

    def filter(self, expression=None, **options):
        return self.chain(Filter(self.queryable, expression, options))

    def recurse(self, column):
        # TODO: This is clearly wrong. It mutates the query, possibly
        # effecting other uses.
        assert isinstance(self.queryable, Select)
        self.queryable.recurse_column = column
        return self

    def limit(self, limit):
        # TODO: This is clearly wrong. It mutates the query, possibly
        # effecting other uses.
        assert isinstance(self.queryable, Select)
        self.queryable.limit = limit
        return self

    def order(self, columns):
        # TODO: This is clearly wrong. It mutates the query, possibly
        # effecting other uses.
        assert isinstance(self.queryable, Select)
        self.queryable.order = columns
        return self

    def join(self, separator=' '):
        return self.chain(
            Join(self.queryable, separator)
        )

    def branch(self, query):
        if self.queryable is not None:
            self.branches.append(self.queryable)
        self.branches.extend(query.branches)
        return self.chain(query.queryable)

    def parse_value_dict(self, queryable, values):
        parsed_values = {}
        for column_name, value in values.items():
            if isinstance(value, Q):
                queryable = self.merge_query_branches(value)
                value = queryable
            parsed_values[column_name] = self.parse_value(queryable, column_name, value)
        return parsed_values

    def parse_bulk_values(self, columns, values):
        return Values(values, columns)

    def parse_value(self, queryable, column_name, value):
        return value

    def get_mutation_source(self):
        if self.queryable.mutatable:
            source = self.queryable
        else:
            source = self.queryable.root
            self.split_branch()
        return source

    def split_branch(self):
        if self.should_branch():
            self.branches.append(self.queryable)

    def should_branch(self):
        return isinstance(self.queryable, Query)

    def iter_branches(self):
        for branch in self.branches:
            yield branch

    def chain(self, queryable):
        return self.__class__(queryable, self.branches, session=self.session)

    def merge_query_branches(self, query):
        queryable = query.queryable
        self.branches = (self.branches or []) + (query.branches or [])
        return queryable



import inspect

from psycopg2.sql import SQL, Identifier

from ..model import field as mf
from ..utils.predicates import not_empty

__all__ = (
    'db_field_registry', 'get_db_field',
    'TextField', 'IntField',
    'RelatedField'
)

db_field_registry = {
}


def get_db_field(field):
    # TODO: Could cache the results against the source field?
    for base_class in inspect.getmro(field.__class__):
        candidates = db_field_registry.get(base_class)
        if candidates:
            for c in candidates:
                if c.match(field):
                    return c(field)
    raise Exception(f'unknown database field type {field}')


class FieldMetaclass(type):
    def __init__(cls, name, bases, dct):
        if bases and name != 'Field':
            for src in getattr(cls, 'sources', ()):
                db_field_registry.setdefault(src, []).append(cls)
        super().__init__(name, bases, dct)


class Field(metaclass=FieldMetaclass):
    db_type = None
    sources = ()

    @classmethod
    def match(cls, field):
        return True

    def __init__(self, source):
        self.source = source

    @property
    def name(self):
        return self.source.name

    @property
    def type(self):
        return self.db_type

    @property
    def unique(self):
        return self.source.unique

    @property
    def null(self):
        return self.source.null

    @property
    def primary_key(self):
        return getattr(self.source, 'primary_key', None)

    def is_concrete(self):
        return True

    def get_value(self, model):
        return getattr(model, self.name, None)

    def get_create_sql(self):
        return SQL(' ').join(filter(not_empty, (
            Identifier(self.name),
            self.get_type_sql(),
            self.get_constraints_sql()
        )))

    def get_type_sql(self):
        return SQL(self.type)

    def get_constraints_sql(self, extra=()):
        src = self.source
        elems = tuple(filter(not_empty, (
            SQL('NOT NULL') if not src.null else None,
            SQL('UNIQUE') if src.unique else None,
            SQL('DEFAULT {}').format(self.translate_default(src.default))
            if src.default is not None else None,
            *extra
        )))
        if not elems:
            return None
        return SQL(' ').join(elems)

    def translate_default(self, default):
        return default


class TextField(Field):
    db_type = 'text'
    sources = (mf.TextField,)

    @property
    def type(self):
        if self.length:
            # TODO: Validate length.
            return f'varchar({self.length})'
        else:
            return super().type

    @property
    def length(self):
        return self.source.length


class IntField(Field):
    db_type = 'int'
    sources = (mf.IntField,)

    @property
    def type(self):
        if self.primary_key:
            return 'serial'
        else:
            return super().type

    def get_constraints_sql(self, extra=()):
        return super().get_constraints_sql((
            SQL('PRIMARY KEY') if self.primary_key else None,
            *extra
        ))


class PasswordField(TextField):
    db_type = 'chkpass'
    sources = (mf.PasswordField,)


class UUIDField(Field):
    db_type = 'uuid'
    sources = (mf.UUIDField,)

    def translate_default(self, default):
        if self.auto:
            return 'uuid_generate_v4()'
        else:
            return default


class PointField(Field):
    db_type = 'point'
    sources = (mf.PointField,)


class DateField(Field):
    db_type = 'date'
    sources = (mf.DateField,)


class DatetimeField(Field):
    db_type = 'timestamptz'
    sources = (mf.DatetimeField,)


class RelatedField(IntField):
    sources = (mf.RelatedField,)

    @property
    def references(self):
        other_table = self.source.other.Meta.table_name
        result = f'{other_table}.id'
        app = self.source.other.Meta.app
        if app:
            result = f'{app.name}.{result}'
        return result

    # def get_other(self):
    #     # TODO: Cache this?
    #     return get_db_field(self.source.other)

    # def get_type_sql(self):
    #     return self.get_other().get_type_sql()

    def get_value(self, model):
        value = super().get_value(model)
        if value:
            # TODO: Shouldn't be named `id`, should be whatever the
            # linked field is. Also, how to handle an integer or
            # whatever value the PK is?
            value = value.id
        return value

    def get_constraints_sql(self):
        other_table = self.source.other.Meta.table_name
        return super().get_constraints_sql((
            SQL('REFERENCES {}(id)').format(Identifier(other_table)),
        ))


class ReverseField(IntField):
    sources = (mf.ReverseField,)

    def is_concrete(self):
        return False



import click

from ..cli.feedback import HaloFeedback
from ..core.registry import Registry, RegistryMetaclass
from ..utils.stringcase import snakecase

Registry('command_registry')


class CommandMetaclass(RegistryMetaclass):
    def __init__(cls, name, bases, dct):
        if not cls.is_baseclass(name, bases):
            command_name = getattr(cls, 'name', snakecase(name))
            cls.name = command_name
        super().__init__(name, bases, dct)
        if not cls.is_baseclass(name, bases):
            from ..cli.admin import admin  # TODO: Ugh.
            params = list(cls().get_params() or ())
            command = click.Command(name=cls.name, params=params, callback=cls._run)
            admin.add_command(command)


class Command(metaclass=CommandMetaclass):
    _registry = 'command_registry'
    _registry_base = 'Command'
    Argument = click.Argument
    Option = click.Option

    def __init__(self, project=None):
        self.project = project

    @classmethod
    def _run(cls, *args, **kwargs):
        ctx = click.get_current_context()
        project = ctx.obj.get('project')
        deployment = ctx.obj.get('deployment')
        # TODO: Support running on all deployments?
        if project and deployment:
            cls._run_remote(project, deployment, *args, **kwargs)
            return
        from ..project.project import load_project  # TODO: Ugh.
        project = load_project()
        project.prepare()  # TODO: Need this?
        command = cls(project)
        command.run(*args, **kwargs)

    @classmethod
    def _run_remote(cls, project, deployment, *args, **kwargs):
        # TODO: Obvs this is too restrictive. We'll need to be able to
        # select the appropriate deployment backend.
        from ..deploy.aws.admin import run_command as aws_run_command
        aws_run_command(project, deployment, cls.name, args, kwargs, feedback=HaloFeedback())

    def get_options(cls):
        pass

    def run(self):
        raise NotImplementedError



from pathlib import Path

from polecat.core.context import active_context

from ..decorators import dbcursor
from ..schema import Schema
from .differ import Differ
from .migration import Migration
from .utils import project_migrations_path


def diff_schemas(to_schema, from_schema=None):
    if from_schema is None:
        from_schema = Schema()
    return Differ().diff(from_schema, to_schema)


@dbcursor
def migrate(migration_paths=None, apps=None, cursor=None):
    bootstrap_migrations()
    migrations = load_migrations(migration_paths, apps=apps)
    schema = Schema()
    for migration in migrations.values():
        migration.forward(schema, migrations, cursor=cursor)
    for migration in migrations.values():
        migration.forward_post_ops(cursor=cursor)


def load_migrations(migration_paths=None, apps=None):
    migrations = {}
    for app in apps or ():
        migrations.update(load_app_migrations(app))
    for path in migration_paths or ():
        migrations.update(load_path_migrations(path))
    migrations.update(load_path_migrations(project_migrations_path()))
    return migrations


@dbcursor
def bootstrap_migrations(cursor):
    # TODO: This is only for development purposes, and *should* fail
    # in production as production users shouldn't have superuser
    # access.
    try:
        sql = (
            'CREATE EXTENSION IF NOT EXISTS chkpass;'
            'CREATE EXTENSION IF NOT EXISTS "uuid-ossp";'
            'CREATE EXTENSION IF NOT EXISTS pg_trgm;'
        )
        cursor.execute(sql)
    except Exception:
        # TODO: Reduce this to programming error.
        pass
    sql = (
        'CREATE TABLE IF NOT EXISTS polecat_migrations ('
        '  id serial primary key,'
        '  app varchar(256),'
        '  name text,'
        '  applied timestamptz'
        ');'
    )
    cursor.execute(sql)


def load_app_migrations(app):
    migrations = {}
    try:
        for path in (app.path / 'migrations').iterdir():
            match = Migration.filename_prog.match(path.name)
            if match:
                migration_class = Migration.load_migration_class(path)
                migration = migration_class(app=app.name, name=path.name[:-3], filename=path.name)
                migrations[f'{app.name}.{path.name[:-3]}'] = migration
                migration.set_app(app)
    except FileNotFoundError:
        pass
    return migrations


def load_path_migrations(path):
    migrations = {}
    try:
        for file_path in (Path(path) / 'migrations').iterdir():
            match = Migration.filename_prog.match(file_path.name)
            if match:
                migration_class = Migration.load_migration_class(file_path)
                migration = migration_class(name=file_path.name[:-3])
                migrations[f'{file_path.name[:-3]}'] = migration
    except FileNotFoundError:
        pass
    return migrations


def make_migrations(to_schema=None, output_path=None, apps=None, migration_paths=None):
    from_schema = Schema()
    migrations = load_migrations(apps=apps, migration_paths=migration_paths)
    for mgr in migrations.values():
        mgr.forward(schema=from_schema, migrations=migrations, schema_only=True)
    if to_schema is None:
        to_schema = active_context().db.schema
    new_migrations = diff_schemas(to_schema, from_schema)
    for mgr in new_migrations:
        # TODO: Feedback?
        mgr.save(output_path=output_path)
    return new_migrations


def make_migrations_for_app(app):
    for role in app.roles:
        make_migrations_for_role(role)
    for model in app.models:
        make_migrations_for_model(model)


def make_migrations_for_role(role):
    pass


def make_migrations_for_model(model):
    pass



class Session:
    def __init__(self, role=None, variables=None):
        self.role = role
        self.variables = variables or {}

    def is_empty(self):
        return not self.role and not self.variables
